import keras
from teras._src.api_export import teras_export


@teras_export("teras.layers.SAINTProjectionHead")
class SAINTProjectionHead(keras.layers.Layer):
    """
    Projection Head layer that is used in the contrastive learning phase of
    the `SAINTPretrainer` to project embeddings to a lower dimension.
    According to the SAINT paper,
    "The use of a projection head to reduce dimensionality before computing
    contrastive loss is common in vision and indeed also improves results
    on tabular data."

    Reference(s):
    https://arxiv.org/abs/2106.01342

    Args:
        hidden_dim: int, Dimensionality of the hidden layer.
            In the official implementation, it is computed as follows,
            `hidden_dim = 6 * embedding_dim * number_of_features // 5`
        output_dim: int, Dimensionality of the output layer.
            In the official implementation, it is computed as follows,
            `output_dim = embedding_dim * number_of_features // 2`
        hidden_activation: Activation function to use in the hidden layer.
            Defaults to "relu".
    """
    def __init__(self,
                 hidden_dim: int,
                 output_dim: int,
                 hidden_activation="relu",
                 **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.hidden_activation = hidden_activation
        self.output_dim = output_dim

        self.hidden_block = keras.layers.Dense(
            units=hidden_dim,
            activation=hidden_activation,
            name="projection_head_hidden"
        )
        self.output_layer = keras.layers.Dense(
            units=output_dim,
            name="projection_head_output"
        )

    def build(self, input_shape):
        self.hidden_block.build(input_shape)
        input_shape = self.hidden_block.compute_output_shape(input_shape)
        self.output_layer.build(input_shape)

    def call(self, inputs):
        x = self.hidden_block(inputs)
        return self.output_layer(x)

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.output_dim,)

    def get_config(self):
        config = {
            "name": self.name,
            "trainable": self.trainable,
            "hidden_dim": self.hidden_dim,
            "output_dim": self.output_dim,
            "hidden_activation": self.hidden_activation,
        }
        return config



import keras
from teras._src.api_export import teras_export


@teras_export("teras.tasks.Imputer")
class Imputer:
    """
    Imputer task class used to impute missing data using the trained `model`
    instance.

    Args:
        model: keras.Model, trained instance of a keras model that will be
            used to impute missing data.
            Currently, `teras` offers `GAIN` and `PCGAIN` architectures for
            imputation.
        data_transformer: Instance of a relevant data transformer that is
            used during the data transformation step before training the
            respective architecture.
    """
    def __init__(self,
                 model: keras.Model,
                 data_transformer=None):
        self.model = model
        self.data_transformer = data_transformer

    def impute(self, x, reverse_transform=True, batch_size=None,
               verbose="auto", steps=None, callbacks=None):
        """
        Imputes the missing data.
        It exposes all the arguments taken by the `predict` method.

        Args:
            x: pd.DataFrame, dataset with missing values.
            reverse_transform: bool, default True, whether to reverse
                transformed the raw imputed data to original format.

        Returns:
            Imputed data in the original format.
        """
        if reverse_transform:
            if self.data_transformer is None:
                raise ValueError(
                    "To reverse transform the raw imputed data, "
                    "`data_transformer` must not be None. "
                    "Please pass the instance of `DataTransformer` class used "
                    "to transform the input data as argument to this `impute` "
                    "method. \n"
                    "Or alternatively, you can set `reverse_transform` "
                    "parameter to False, and manually reverse transform the "
                    "generated raw data to original format using the "
                    "`reverse_transform` method of `DataTransformer` instance.")
        x_imputed = self.model.predict(x,
                                       batch_size=batch_size,
                                       verbose=verbose,
                                       steps=steps,
                                       callbacks=callbacks)
        if reverse_transform:
            x_imputed = self.data_transformer.reverse_transform(x_imputed)
        return x_imputed



import keras

from teras._src.layers.cls_token import CLSToken
from teras._src.layers.cls_token_extraction import CLSTokenExtraction
from teras._src.layers.saint.embedding import SAINTEmbedding
from teras._src.layers.saint.encoder_layer import SAINTEncoderLayer
from teras._src.models.backbones.backbone import Backbone
from teras._src.api_export import teras_export


@teras_export("teras.model.SAINTBackbone")
class SAINTBackbone(Backbone):
    """
    SAINT Backbone based on the SAINT architecture proposed in the paper,
    "SAINT: Improved Neural Networks for Tabular Data".

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        input_dim: int, dimensionality of the input dataset. i.e. the
            number of features in the dataset.
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use any value <=0 as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings used
            by the model. It is also referred to as the `d_model` or
            model dimensionality.
        num_layers: int, number of `TransformerEncoderLayer`s to use in
            the encoder.
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
            `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.
        embedd_inputs: bool, whether to use `SAINTEmbedding` layer to
            create emebddings of inputs. Defaults to `True` as this is
            what we want. but when pretraining we want to use the
            `SAINTBackbone` as encoder only which expects embeddings as
            inputs, so that's when we set this parameter to `False`.
        return_cls_token_only: bool, whether to return only embeddings
            for the `CLS` token. Defaults to `True`.
    """
    def __init__(self,
                 input_dim: int,
                 cardinalities: list,
                 embedding_dim: int,
                 num_layers: int = 6,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 embedd_inputs: bool = True,
                 return_cls_token_only: bool = True,
                 **kwargs):
        if embedd_inputs:
            inputs = keras.layers.Input((input_dim,), name="inputs")
        else:
            # Embeddings will be computed by an external embedding layer,
            # so in that case, the model should expect inputs with dims
            # (input_dim, embedding_dim)
            inputs = keras.layers.Input((input_dim, embedding_dim),
                                        name="inputs")
        x = inputs
        if embedd_inputs:
            x = SAINTEmbedding(
                embedding_dim=embedding_dim,
                cardinalities=cardinalities)(x)
        x = CLSToken(embedding_dim=embedding_dim)(x)
        for i in range(num_layers):
            x = SAINTEncoderLayer(
                embedding_dim=embedding_dim,
                num_heads=num_heads,
                feedforward_dim=feedforward_dim,
                attention_dropout=attention_dropout,
                feedforward_dropout=feedforward_dropout,
                layer_norm_epsilon=layer_norm_epsilon,
                name=f"saint_encoder_layer_{i}")(x)
        if return_cls_token_only:
            outputs = CLSTokenExtraction()(x)
        else:
            outputs = x
        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        self.input_dim = input_dim
        self.cardinalities = cardinalities
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon
        self.embedd_inputs = embedd_inputs
        self.return_cls_token_only = return_cls_token_only

    def compute_output_shape(self, input_shape):
        if self.return_cls_token_only:
            return input_shape[:1] + (1, self.embedding_dim)
        else:
            return input_shape + (self.embedding_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "input_dim": self.input_dim,
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim,
            "num_layers": self.num_layers,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon,
            "embedd_inputs": self.embedd_inputs,
            "return_cls_token_only": self.return_cls_token_only,
        })
        return config



from keras import random, ops
from teras._src.layers.tabnet.feature_transformer_layer import TabNetFeatureTransformerLayer
from keras.src.testing.test_case import TestCase


class TabNetFeatureTransformerLayerTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.input_batch = random.normal((self.batch_size, 5))

    def test_valid_output_shape(self):
        layer = TabNetFeatureTransformerLayer(dim=8,
                                              batch_momentum=0.9)
        outputs = layer(self.input_batch)
        self.assertEqual(ops.shape(outputs), (self.batch_size, 8))




import torch
from keras import random, ops
from teras._src.backend.common.models.gans.ctgan.discriminator import BaseCTGANDiscriminator
from teras._src.typing import IntegerSequence


class CTGANDiscriminator(BaseCTGANDiscriminator):
    def __init__(self,
                 hidden_dims: IntegerSequence = (256, 256),
                 packing_degree: int = 8,
                 gradient_penalty_lambda: float = 10.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(hidden_dims=hidden_dims,
                         packing_degree=packing_degree,
                         gradient_penalty_lambda=gradient_penalty_lambda,
                         seed=seed,
                         **kwargs)

    def gradient_penalty(self,
                         real_samples,
                         generated_samples):
        self.zero_grad()
        batch_size = ops.shape(real_samples)[0]
        dim = ops.shape(real_samples)[1]

        alpha = random.uniform(
            shape=(batch_size // self.packing_degree, 1, 1),
            seed=self.seed_gen,
        )
        alpha = ops.reshape(
            ops.tile(alpha, [1, self.packing_degree, dim]),
            (-1, dim))
        interpolated_samples = ((alpha * real_samples)
                                + ((1 - alpha) * generated_samples))
        interpolated_samples.requires_grad = True
        y_interpolated = self(interpolated_samples)
        y_interpolated.backward()
        gradients = interpolated_samples.grad
        gradients = ops.reshape(gradients,
                                newshape=(-1, self.packing_degree * dim))
        # Calculating gradient penalty
        gradients_norm = torch.norm(gradients)

        gradient_penalty = (
                ops.mean(ops.square(gradients_norm - 1.0)) *
                self.gradient_penalty_lambda
        )
        return gradient_penalty



import torch
import keras
from teras._src.backend.common.models.pretrainers.tab_transformer import (
    BaseTabTransformerMLMPretrainer,
    BaseTabTransformerRTDPretrainer
)
from keras import ops


class TabTransformerMLMPretrainer(BaseTabTransformerMLMPretrainer):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 missing_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         missing_rate=missing_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def train_step(self, data):
        mask = self._create_mask(ops.shape(data))

        # Call torch.nn.Module.zero_grad() to clear the leftover gradients
        # for the weights from the previous train step.
        self.zero_grad()

        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        y_pred = self(data, (1 - mask))
        loss = self.compute_loss(y=data * mask,
                                 y_pred=y_pred)
        # Run backwards pass
        loss.backward()
        # Get grads
        gradients = [v.value.grad for v in self.trainable_variables]
        # Optimize
        with torch.no_grad():
            self.optimizer.apply(
                gradients,
                self.trainable_variables
            )

        logs = {m.name: m.result() for m in self.metrics}
        return logs


class TabTransformerRTDPretrainer(BaseTabTransformerRTDPretrainer):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 replace_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         replace_rate=replace_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def train_step(self, data):
        mask = self._create_mask(ops.shape(data))
        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        y_pred = self(data, mask=(1 - mask))
        loss = self.compute_loss(y=mask,
                                 y_pred=y_pred)
        # Run backward pass
        loss.backward()
        # Get grads
        gradients = [v.value.grad for v in self.trainable_variables]
        # Optimize
        self.optimizer.apply(
            gradients,
            self.trainable_variables
        )

        logs = {m.name: m.result() for m in self.metrics}
        return logs



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.CategoricalEmbedding")
class CategoricalEmbedding(keras.layers.Layer):
    """
    Categorical Embedding layer that create trainable embeddings for
    categorical features values.

    Args:
        embedding_dim: int, dimensionality of the embeddings
        cardinalities: list or ndarray, a list or 1d-array of
            cardinalities of all the features in the dataset in the
            same order as the features' occurrence.
            For numerical features, use 0 as indicator at
            the corresponding index of the array.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
    """
    def __init__(self,
                 embedding_dim: int,
                 cardinalities: list,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.cardinalities = cardinalities
        self.embedding_layers = []
        self._categorical_idx = []
        for idx, card in enumerate(self.cardinalities):
            if card > 0:
                embedding = keras.layers.Embedding(
                    input_dim=card + 1,
                    output_dim=self.embedding_dim)
                self.embedding_layers.append(embedding)
                self._categorical_idx.append(idx)

    def call(self, inputs):
        idx = self._categorical_idx[0]
        feature = ops.expand_dims(inputs[:, idx], axis=1)
        categorical_embeddings = self.embedding_layers[0](feature)
        for feature_idx, embedding_layer in zip(self._categorical_idx[1:],
                                                self.embedding_layers[1:]):
            feature = ops.expand_dims(inputs[:, feature_idx], axis=1)
            f_e = embedding_layer(feature)
            # Concatenate feature embeddings along the feature axis
            categorical_embeddings = ops.concatenate(
                xs=[categorical_embeddings, f_e], axis=1)
        return categorical_embeddings

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (len(self._categorical_idx),
                                   self.embedding_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "embedding_dim": self.embedding_dim,
            "cardinalities": self.cardinalities
        })



def assert_fitted(func):
    def wrapper(*args, **kwargs):
        self, *_ = args
        if not self._fitted:
            raise AssertionError(
                f"`{self.__dict__.__class__}` has not been fitted yet. "
                f"Please fit by calling the `fit` method before accessing the "
                f"{func.__name__} method.")
        return func(*args, **kwargs)
    return wrapper



import os

import keras
from keras import ops
from keras.src.testing.test_case import TestCase

from teras._src.models.gans.ctgan.discriminator import CTGANDiscriminator


class CTGANDiscriminatorTest(TestCase):
    def setUp(self):
        self.inputs = ops.ones((16, 5))
        self.packing_degree = 4

    def test_valid_call(self):
        discriminator = CTGANDiscriminator(packing_degree=self.packing_degree)
        outputs = discriminator(self.inputs)

    def test_valid_output_shape(self):
        discriminator = CTGANDiscriminator(packing_degree=self.packing_degree)
        outputs = discriminator(self.inputs)
        self.assertEqual(ops.shape(outputs), (16 // self.packing_degree, 1))

    def test_save_and_load(self):
        discriminator = CTGANDiscriminator(packing_degree=self.packing_degree)
        outputs_original = discriminator(self.inputs)
        save_path = os.path.join(self.get_temp_dir(),
                                 "ctgan_discriminator.keras")
        discriminator.save(save_path)
        reloaded_discriminator = keras.models.load_model(save_path)
        self.assertIsInstance(reloaded_discriminator, CTGANDiscriminator)
        outputs_reloaded = reloaded_discriminator(self.inputs)
        self.assertAllClose(ops.convert_to_numpy(outputs_original),
                            ops.convert_to_numpy(outputs_reloaded))



import keras

from teras._src.layers.transformer.encoder_layer import TransformerEncoderLayer
from teras._src.models.backbones.backbone import Backbone
from teras._src.api_export import teras_export


@teras_export("teras.models.TransformerEncoderBackbone")
class TransformerEncoderBackbone(Backbone):
    """
    Transformer Encoder model as proposed in the
    "Attention is all you need" paper.

    Reference(s):
        https://arxiv.org/abs/1706.03762

    Args:
        input_dim: int, dimensionality of the input data.
        embedding_dim: int, dimensionality of the embeddings used
            by the model. It is also referred to as the `d_model` or
            model dimensionality.
        num_layers: int, number of `TransformerEncoderLayer`s to use in
            the encoder.
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
        `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.
        unnormalized_layers: list, list of indices corresponding to
            the layers in which `LayerNormalization` won't be used.
            For instance, if you don't want to use the normalization in
            the first `TransformerEncoderLayer` layer (like FT-Transformer)
            you can pass [0]. If you don't want to normalize first and
            second layer, you can similarly pass [0, 1] and so on.
            Defaults, to `[]` (empty list), because the original
            Transformer architecture and most others use normalization
            in all of their layers.
        pre_normalization: bool, whether to use Pre-Normalization technique
            whereby `LayerNormalization` is applied to inputs of the
            `MultiHeadAttention` or `FeedForward` and then outputs of
            those layers are elementwise added to the original inputs.
            Defaults to `False`, as the original Transformers architecture
            doesn't use pre-normalization.
    """
    def __init__(self,
                 input_dim: int,
                 embedding_dim: int,
                 num_layers: int = 6,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 unnormalized_layers: list = [],
                 pre_normalization: bool = False,
                 **kwargs):
        if num_layers < 1:
            raise ValueError(
                f"`num_layers` must be 1 or greater. Received {num_layers}")
        if (len(unnormalized_layers) > 0 and
                (max(unnormalized_layers) > (num_layers - 1) or
                 min(unnormalized_layers) < 0)):
            raise ValueError(
                f"Layer indices must be in the interval [0, num_layers). "
                f"Received {unnormalized_layers}"
            )
        inputs = keras.layers.Input(shape=(input_dim, embedding_dim))
        x = inputs
        for i in range(num_layers):
            use_normalization = i not in unnormalized_layers
            x = TransformerEncoderLayer(
                embedding_dim=embedding_dim,
                num_heads=num_heads,
                feedforward_dim=feedforward_dim,
                attention_dropout=attention_dropout,
                feedforward_dropout=feedforward_dropout,
                layer_norm_epsilon=layer_norm_epsilon,
                use_normalization=use_normalization,
                pre_normalization=pre_normalization,
                name=f"transformer_encoder_layer_{i+1}")(x)
        outputs = x
        super().__init__(inputs=inputs, outputs=outputs, **kwargs)
        self.input_dim = input_dim
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon
        self.unnormalized_layers = unnormalized_layers
        self.pre_normalization = pre_normalization

    def get_config(self):
        config = super().get_config()
        config.update({
            "input_dim": self.input_dim,
            "embedding_dim": self.embedding_dim,
            "num_layers": self.num_layers,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon,
            "unnormalized_layers": self.unnormalized_layers,
            "pre_normalization": self.pre_normalization
        })
        return config



import jax
import keras
from keras.models import Model
from keras.src.saving import serialization_lib

# TODO:
#   1. Implement a compose state function
#   2. Also maybe replace _parse_variables with decompose state


class JAXGAN(Model):
    """
    Extends Keras.Model to support multiple optimizers for GAN based models.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 **kwargs):
        Model.__init__(self, **kwargs)
        self.generator = generator
        self.discriminator = discriminator

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(
            name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(
            name="discriminator_loss")

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(),
                discriminator_optimizer=keras.optimizers.Adam(),
                **kwargs
                ):
        super().compile(**kwargs)
        self.generator_optimizer = keras.optimizers.get(
            generator_optimizer)
        self.discriminator_optimizer = keras.optimizers.get(
            discriminator_optimizer)

    def get_compile_config(self):
        config = super().get_compile_config()
        config.update(
            serialization_lib.SerializableDict(
                generator_optimizer=self.generator_optimizer,
                discriminator_optimizer=self.discriminator_optimizer,
            )
        )
        return config.serialize()

    def compile_from_config(self, config):
        config = serialization_lib.deserialize_keras_object(config)
        generator_optimizer = config.pop("generator_optimizer")
        discriminator_optimizer = config.pop("discriminator_optimizer")
        self.compile(generator_optimizer=generator_optimizer,
                     discriminator_optimizer=discriminator_optimizer,
                     **config)
        if self.built:
            self.generator_optimizer.build(self.generator.trainable_variables)
            self.discriminator_optimizer.build(
                self.discriminator.trainable_variables)

    @property
    def optimizers(self):
        return [self.generator_optimizer, self.discriminator_optimizer]

    @property
    def optimizers_variables(self):
        vars = []
        for optimizer in self.optimizers:
            vars.extend(optimizer.variables)
        return vars

    def build_optimizers(self):
        if not self.built:
            raise AssertionError(
                "You must build the model before calling `build_optimizers()`."
            )
        self.generator_optimizer.build(self.generator.trainable_variables)
        self.discriminator_optimizer.build(
            self.discriminator.trainable_variables)

    @property
    def metrics(self):
        metrics = [
            self.generator_loss_tracker,
            self.discriminator_loss_tracker,
        ]
        return metrics

    # TODO: Good idea (?) but needs more work to fix!
    def _parse_state(self, state):
        """
        Decomposes overall gan model variables into generator and
        discriminator's variables.

        Returns:
            generator_state, discriminator_state, other_state, metrics_vars
        """
        generator_state = []
        discriminator_state = []
        other_state = []    # sometimes the parent class may have its own vars
        state = list(state)
        while (len(state) < 4):
            # we append empty dummy placeholder lists
            state.append([])
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        )  = state
        # Get generator state
        # Since generator comes and gets built before discriminator
        num_gen_train_vars = len(self.generator.trainable_variables)
        generator_state.append(
            trainable_variables[:num_gen_train_vars])
        num_gen_non_train_vars = len(self.generator.non_trainable_variables)
        generator_state.append(
            non_trainable_variables[:num_gen_non_train_vars])
        if optimizer_variables is not None:
            generator_state.append(
                optimizer_variables[:len(self.generator_optimizer.variables)])

        # Get discriminator state
        num_disc_train_vars = len(self.discriminator.trainable_variables)
        slice_size = num_gen_train_vars + num_disc_train_vars
        discriminator_state.append(
            trainable_variables[num_gen_train_vars: slice_size])
        num_disc_non_train_vars = len(
            self.discriminator.non_trainable_variables)
        slice_size = num_gen_non_train_vars + num_disc_non_train_vars
        discriminator_state.append(
            non_trainable_variables[num_gen_non_train_vars: slice_size])
        if optimizer_variables is not None:
            discriminator_state.append(
                optimizer_variables[len(self.generator_optimizer.variables):])

        # Get anything left over
        other_state.append(
            trainable_variables[num_gen_train_vars + num_disc_train_vars:]
        )
        other_state.append(
            trainable_variables[num_gen_non_train_vars +
                                num_disc_non_train_vars:]
        )

        return (tuple(generator_state), tuple(discriminator_state),
                tuple(other_state), metrics_variables)

    def _parse_variables(self, trainable_variables, non_trainable_variables,
                         optimizer_variables=None):
        """
        Decomposes overall gan model variables into generator and
        discriminator's variables.
        """
        # Get generator state
        generator_state = []
        # Since generator comes and gets built before discriminator
        num_gen_train_vars = len(self.generator.trainable_variables)
        generator_state.append(
            trainable_variables[:num_gen_train_vars])
        num_gen_non_train_vars = len(self.generator.non_trainable_variables)
        generator_state.append(
            non_trainable_variables[:num_gen_non_train_vars])
        if optimizer_variables is not None:
            generator_state.append(
                optimizer_variables[:len(self.generator_optimizer.variables)])

        # Get discriminator state
        discriminator_state = []
        slice_size = num_gen_train_vars + len(self.discriminator.trainable_variables)
        discriminator_state.append(
            trainable_variables[num_gen_train_vars: slice_size])
        slice_size = num_gen_non_train_vars + len(
            self.discriminator.non_trainable_variables)
        discriminator_state.append(
            non_trainable_variables[num_gen_non_train_vars: slice_size])
        if optimizer_variables is not None:
            discriminator_state.append(
                optimizer_variables[len(self.generator_optimizer.variables):])
        return tuple(generator_state), tuple(discriminator_state)

    def compute_loss(self, **kwargs):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` doesn't provide an implementation for"
            f" the `compute_loss` method. Please use "
            f"`compute_discriminator_loss` or `compute_generator_loss` for "
            f"relevant purpose."
        )

    def call(self, **kwargs):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` doesn't provide an implementation "
            f"for the `call` method. Please use the call method of "
            f"`GAIN().generator` or `GAIN().discriminator`."
        )

    def get_generator(self):
        return self.generator

    def get_discriminator(self):
        return self.discriminator

    def get_config(self):
        config = super().get_config()
        config.update({
            'generator': keras.layers.serialize(self.generator),
            'discriminator': keras.layers.serialize(self.discriminator),
            'hint_rate': self.hint_rate,
            'alpha': self.alpha,
        })
        return config

    @classmethod
    def from_config(cls, config):
        generator = keras.layers.deserialize(config.pop("generator"))
        discriminator = keras.layers.deserialize(config.pop("discriminator"))
        return cls(generator=generator, discriminator=discriminator,
                   **config)

    def _get_jax_state(
            self,
            trainable_variables=False,
            non_trainable_variables=False,
            optimizer_variables=False,
            metrics_variables=False,
            purge_model_variables=False,
    ):
        state = []
        if trainable_variables:
            state.append([v.value for v in self.trainable_variables])
        if non_trainable_variables:
            state.append([v.value for v in self.non_trainable_variables])
        if optimizer_variables:
            state.append([v.value for v in self.optimizers_variables])
        if metrics_variables:
            state.append([v.value for v in self.metrics_variables])
        if purge_model_variables:
            self._purge_model_variables(
                trainable_variables=trainable_variables,
                non_trainable_variables=non_trainable_variables,
                optimizer_variables=optimizer_variables,
                metrics_variables=metrics_variables,
            )
        return tuple(state)

    def jax_state_sync(self):
        if not getattr(self, "_jax_state", None) or self._jax_state_synced:
            return

        trainable_variables = self._jax_state.get("trainable_variables", None)
        non_trainable_variables = self._jax_state.get(
            "non_trainable_variables", None
        )
        optimizer_variables = self._jax_state.get("optimizer_variables", None)
        metrics_variables = self._jax_state.get("metrics_variables", None)
        if trainable_variables:
            for ref_v, v in zip(self.trainable_variables, trainable_variables):
                ref_v.assign(v)
        if non_trainable_variables:
            for ref_v, v in zip(
                    self.non_trainable_variables, non_trainable_variables
            ):
                ref_v.assign(v)
        if optimizer_variables:
            for ref_v, v in zip(self.optimizers_variables, optimizer_variables):
                ref_v.assign(v)
        if metrics_variables:
            for ref_v, v in zip(self.metrics_variables, metrics_variables):
                ref_v.assign(v)
        self._jax_state_synced = True

    def _record_training_state_sharding_spec(self):
        self._trainable_variable_shardings = [
            v.value.sharding for v in self.trainable_variables
        ]
        self._non_trainable_variable_shardings = [
            v.value.sharding for v in self.non_trainable_variables
        ]
        self._optimizer_variable_shardings = [
            v.value.sharding for v in self.optimizers_variables
        ]
        self._metrics_variable_shardings = [
            v.value.sharding for v in self.metrics_variables
        ]

    def _enforce_jax_state_sharding(
            self,
            trainable_variables=None,
            non_trainable_variables=None,
            optimizer_variables=None,
            metrics_variables=None,
    ):
        """Enforce the sharding spec constraint for all the training state.

        Since the output of the train/eval step will be used as inputs to next
        step, we need to ensure that they have the same sharding spec, so that
        jax.jit won't have to recompile the train/eval function.

        Note that this function will also rely on the recorded sharding spec
        for each of states.

        This function is expected to be called within the jitted train/eval
        function, especially around the end of the function.
        """
        trainable_variables = trainable_variables or []
        non_trainable_variables = non_trainable_variables or []
        optimizer_variables = optimizer_variables or []
        metrics_variables = metrics_variables or []

        for i in range(len(trainable_variables)):
            trainable_variables[i] = jax.lax.with_sharding_constraint(
                trainable_variables[i], self._trainable_variable_shardings[i]
            )
        for i in range(len(non_trainable_variables)):
            non_trainable_variables[i] = jax.lax.with_sharding_constraint(
                non_trainable_variables[i],
                self._non_trainable_variable_shardings[i],
            )
        for i in range(len(optimizer_variables)):
            optimizer_variables[i] = jax.lax.with_sharding_constraint(
                optimizer_variables[i], self._optimizer_variable_shardings[i]
            )
        for i in range(len(metrics_variables)):
            metrics_variables[i] = jax.lax.with_sharding_constraint(
                metrics_variables[i], self._metrics_variable_shardings[i]
            )
        return (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables,
        )

    def _purge_model_variables(
            self,
            trainable_variables=False,
            non_trainable_variables=False,
            optimizer_variables=False,
            metrics_variables=False,
    ):
        """Remove all the model variable for memory saving.

        During JAX training, since the training function are stateless, we have
        to pass in and get the model weights over and over, during which the
        copy of the weights that attached to the KerasVariable are still and
        occupying extra memory. We remove those variable to save memory (for
        better memory utilization) at the beginning of the epoch, and reattach
        the value back to variables at the end of the epoch, via
        `jax_state_sync()`.
        """
        if trainable_variables:
            for v in self.trainable_variables:
                v._value = None
        if non_trainable_variables:
            for v in self.non_trainable_variables:
                v._value = None
        if optimizer_variables:
            for v in self.optimizer.variables:
                v._value = None
        if metrics_variables:
            for v in self.metrics_variables:
                v._value = None



from teras._src.layers.saint.embedding import SAINTEmbedding
from keras.src.testing.test_case import TestCase
from keras import ops, random


class SAINTEmbeddingTest(TestCase):
    def setUp(self):
        # we assume there are 4 features in the dataset, with feature 1
        # having 5 unique categories, feature 2 having 3 unique
        # categories, feature 3 with 2 unique categories and feature 4
        # with 7 unique categories
        x_cat = ops.array([
            ops.cast(random.randint((16,), minval=0, maxval=card),
                     dtype="float32")
            for card in [5, 3, 2, 7]]
        )
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((16, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous

    def test_valid_output_shape(self):
        embedding = SAINTEmbedding(embedding_dim=8,
                                   cardinalities=self.cardinalities)
        outputs = embedding(self.input_batch)
        self.assertEqual(ops.shape(outputs), (16, 7, 8))



import keras
import numpy as np
from teras._src.api_export import teras_export
from keras import ops
from keras.backend import floatx


@teras_export("teras.layers.FTTransformerFeatureTokenizer")
class FTTransformerFeatureTokenizer(keras.layers.Layer):
    """
    Feature Tokenizer layer based on FT-Transformer architecture
    proposed in the "Revisiting Deep Learning Models for Tabular Data"
    paper.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use any value <=0 as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings

    Shapes:
        Input Shape: `(batch_size, num_features)`
        Output Shape: `(batch_size, num_features, embedding_dim)`
    """
    def __init__(self,
                 cardinalities: list,
                 embedding_dim: int,
                 **kwargs):
        super().__init__(**kwargs)
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim
        self._cardinalities_arr = np.array(cardinalities)
        self._continuous_idx = np.flatnonzero(
            self._cardinalities_arr == 0)
        self._categorical_idx = np.flatnonzero(
            self._cardinalities_arr != 0)

        # we add an extra token for missing value
        num_special_tokens = 1
        self._total_tokens = sum(self.cardinalities) + num_special_tokens
        categorical_cardinalities = self._cardinalities_arr[self._categorical_idx]
        categorical_cardinalities = ops.pad(
            categorical_cardinalities,
            (1, 0),
            constant_values=num_special_tokens,
        )
        self._category_offset = ops.cumsum(categorical_cardinalities)[:-1]
        self._category_offset = ops.cast(self._category_offset, floatx())

    def build(self, input_shape=None):
        self.categorical_embeddings = self.add_weight(
            shape=(self._total_tokens, self.embedding_dim),
            initializer="random_normal",
            trainable=True,
        )
        num_continuous_features = len(self._continuous_idx)
        self.continuous_embeddings = self.add_weight(
            shape=(num_continuous_features,
                   self.embedding_dim),
            trainable=True,
        )
        self.bias = self.add_weight(
            shape=(len(self.cardinalities),),
            initializer="random_normal",
            trainable=True,
        )

    def call(self, inputs):
        categorical = ops.take(inputs,
                               indices=self._categorical_idx,
                               axis=1)
        categorical += self._category_offset
        continuous = ops.take(inputs,
                              indices=self._continuous_idx,
                              axis=1)
        categorical = ops.cast(categorical, "int32")
        categorical = ops.take(self.categorical_embeddings,
                               indices=categorical,
                               axis=0)
        continuous = (ops.expand_dims(continuous, -1) *
                      ops.expand_dims(self.continuous_embeddings, 0))
        out = ops.concatenate([continuous, categorical], axis=1)
        out += ops.reshape(self.bias, (1, ops.shape(self.bias)[0], 1))
        return out

    def compute_output_shape(self, input_shape):
        return input_shape + (self.embedding_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim
        })
        return config



from tensorflow import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.TVAE")
class TVAE(backend.models.TVAE):
    """
    TVAE is a tabular data generation architecture proposed by Lei Xu et al.
    in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        encoder: keras.Model, an instance of `TVAEEncoder`. It encodes or
            compresses the input.
        decoder: keras.Model, an instance of `TVAEDecoder`. It decodes or
            decompresses from latent dimensions to data dimensions.
        latent_dim: int, Dimensionality of the learned latent space.
            Default 128.
        loss_factor: float, Hyperparameter used in the computation of
            `ELBO loss`. It controls how much the cross entropy loss
            contributes to the overall loss. It is directly proportional to
            the cross entropy loss. Defaults to 2.
    """
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 metadata: dict,
                 data_dim: int,
                 latent_dim: int = 128,
                 loss_factor: float = 2.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(encoder=encoder,
                         decoder=decoder,
                         metadata=metadata,
                         data_dim=data_dim,
                         latent_dim=latent_dim,
                         loss_factor=loss_factor,
                         seed=seed,
                         **kwargs)



import os

import keras.models
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.backbones.transformer.encoder import \
    TransformerEncoderBackbone


class TransformerEncoderTest(TestCase):
    def setUp(self):
        _batch_size, _input_dim, _batch_size = 8, 5, 16
        self.input_shape = (_batch_size, _input_dim, _batch_size)
        self.input_batch = random.normal(self.input_shape)

    def test_valid_call(self):
        model = TransformerEncoderBackbone(
            input_dim=self.input_shape[1],
            num_layers=6,
            embedding_dim=self.input_shape[-1],
            num_heads=8,
            feedforward_dim=64)
        outputs = model(self.input_batch)

    def test_valid_output_shape(self):
        model = TransformerEncoderBackbone(
            input_dim=self.input_shape[1],
            num_layers=6,
            embedding_dim=self.input_shape[-1],
            num_heads=8,
            feedforward_dim=64)
        outputs = model(self.input_batch)
        self.assertEqual(self.input_shape, ops.shape(outputs))

    def test_model_save_and_load(self):
        model = TransformerEncoderBackbone(
            input_dim=self.input_shape[1],
            num_layers=6,
            embedding_dim=self.input_shape[-1],
            num_heads=8,
            feedforward_dim=64)
        outputs = model(self.input_batch)
        save_path = os.path.join(self.get_temp_dir(),
                                 "transformer_encoder_backbone.keras"
                                 )
        model.save(save_path)
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TransformerEncoderBackbone)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.input_batch)
        self.assertAllClose(
            ops.convert_to_numpy(outputs),
            ops.convert_to_numpy(reloaded_outputs)
        )



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.TabTransformerMLMPretrainer")
class TabTransformerMLMPretrainer(backend.models.TabTransformerMLMPretrainer):
    """
    Masked Language Modelling (MLM) based Pretrainer for pretraining
    `TabTransformerBackbone` as proposed by Huang et al. in the paper,
    "TabTransformer: Tabular Data Modeling Using Contextual Embeddings".

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        model: keras.Model, instance of `TabTransformerBackbone` to
            pretrain
        data_dim: int, dimensionality of the input dataset
        missing_rate: float, fraction of original features to make missing.
            Must be in the range [0, 1).
            Defaults to 0.3 (or 30%)
        mask_seed: int, seed for generating mask.
            Defaults to 1337
    """
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 missing_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(
            model=model,
            data_dim=data_dim,
            missing_rate=missing_rate,
            mask_seed=mask_seed,
            **kwargs)


@teras_export("teras.models.TabTransformerRTDPretrainer")
class TabTransformerRTDPretrainer(backend.models.TabTransformerRTDPretrainer):
    """
    Replaced Token Detection (RTD) based Pretrainer for pretraining
    `TabTransformerBackbone` as proposed by Huang et al. in the paper,
    "TabTransformer: Tabular Data Modeling Using Contextual Embeddings".

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        model: keras.Model, instance of `TabTransformerBackbone` to
            pretrain
        data_dim: int, dimensionality of the input dataset
        replace_rate: float, fraction of original features to replace.
            Must be in the range [0, 1).
            Defaults to 0.3 (or 30%)
        mask_seed: int, seed for generating mask.
            Defaults to 1337
        shuffle_seed: int, seed for shuffling inputs.
            Defaults to 1999
    """
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 replace_rate: float = 0.3,
                 mask_seed: int = 1337,
                 shuffle_seed: int = 1999,
                 **kwargs):
        super().__init__(
            model=model,
            data_dim=data_dim,
            replace_rate=replace_rate,
            mask_seed=mask_seed,
            shuffle_seed=shuffle_seed,
            **kwargs)



import keras

from teras._src.api_export import teras_export


@teras_export("teras.models.Backbone")
class Backbone(keras.Model):
    """
    Base class for Backbone models.

    Copied from Keras-CV with some modifications.

    Reference(s):
    https://github.com/keras-team/keras-cv/blob/master/keras_cv/models/backbones/backbone.py
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_config(self):
        config = {"name": self.name,
                  "trainable": self.trainable}
        return config

    @classmethod
    def from_config(cls, config):
        # The default `from_config()` for functional models will return a
        # vanilla `keras.Model`. We override it to get a subclass
        # instance back.
        return cls(**config)



from teras._src.layers.cls_token import CLSToken
from keras.src.testing.test_case import TestCase
from keras import ops, random


class CLSTokenTest(TestCase):
    def setUp(self):
        self.input_shape = (16, 5, 8)
        self.input_batch = random.normal(self.input_shape)

    def test_valid_output_shape(self):
        cls_token = CLSToken(embedding_dim=8)
        outputs = cls_token(self.input_batch)
        self.assertEqual(
            ops.shape(outputs),
            (self.input_shape[0],
             self.input_shape[1] + 1,
             self.input_shape[-1]))



try:
    import tensorflow as tf
except:
    raise ImportError(
        "You need tensorflow to use `TVAEataSampler`. "
        "Install it using `pip install tensorflow`"
    )
import numpy as np
from teras._src.preprocessing.data_samplers.ctgan import CTGANDataSampler as _BaseDataSampler
from teras._src.typing import FeaturesNamesType


class TVAEDataSampler(_BaseDataSampler):
    """
    TVAEDataSampler class for `TVAE` architecture.
    It subclasses the `CTGANDataSampler` class from `CTGAN` architecture.

    The two classes share much functionality since `TVAE` and `CTGAN`
    are proposed in the same paper and almost all preprocessing
    for both is same.
    There are, however, are a few differences in the `get_dataset`
    and `generator` methods, hence this new subclassed class.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        metadata: dict, A dictionary of metadata computed during data
            transformation. You can access it from the `.get_metadata()` of
            `TVAEDataTransformer` instance.
        categorical_features: list, List of categorical features names.
            CTGAN requires dataset to have at least one categorical feature,
            if your dataset doesn't contain any categorical features,
            consider using some other generative model.
        continuous_features: list, List of continuous features names
        batch_size: int, default 512,
            Batch size to use for the dataset.
        seed: int, Seed for random ops.
    """
    def __init__(self,
                 metadata: dict,
                 categorical_features: FeaturesNamesType = None,
                 continuous_features: FeaturesNamesType = None,
                 batch_size: int = 512,
                 seed: int = 1337,
                 ):
        super().__init__(metadata=metadata,
                         categorical_features=categorical_features,
                         continuous_features=continuous_features,
                         batch_size=batch_size,
                         seed=seed)

    def get_dataset(self,
                    x_transformed,
                    x_original=None):
        """
        Args:
            x_transformed: Dataset transformed using `TVAEDataTransformer` class
            x_original: Original Dataset - a pandas DataFrame.
                It is used for computing categorical values' probabilities
                for later sampling.
        Returns:
            Returns a tensorflow dataset that utilizes the `generator` method
            to create batches of data. This way user can just pass the dataset
            object to the fit method of the model and each batch generated
            will satisfy all out requirements of sampling
        """
        self.num_samples, self.data_dim = x_transformed.shape
        # adapting the approach from the official implementation
        # to sample evenly across the categories to combat imbalance
        row_idx_raw = [x_original.groupby(feature).groups
                       for feature in self.categorical_features]
        self.row_idx_by_categories = [
            [values.to_list() for values in feat.values()]
            for feat in row_idx_raw]

        dataset = tf.data.Dataset.from_generator(
            self.generator,
            output_signature=(
                tf.TensorSpec(
                    shape=(self.batch_size, tf.shape(x_transformed)[1]),
                    name="data_batch")
                              ),
            args=(x_transformed,),
        )
        return dataset

    def generator(self, x_transformed):
        """
        Used to create a tensorflow dataset.

        Args:
            x_transformed: Dataset transformed by the `TVAEDataTransformer`
                class.

        Returns:
            A batch of data
        """
        num_steps_per_epoch = self.num_samples // self.batch_size
        for _ in range(num_steps_per_epoch):
            selected_cat_features_idx = self._np_rng.choice(
                np.arange(self._num_categorical_features),
                size=self.batch_size
            )
            # NOTE: We've precomputed the probabilities in the DataTransformer
            # class for each feature already to speed things up.
            selected_features_categories_probs = self._features_categories_probs[selected_cat_features_idx]

            # Choose random values idx for features
            selected_cat_values_idx = np.array([
                self._np_rng.choice(np.arange(len(probs)),
                                    p=probs)
                for probs in selected_features_categories_probs]
            ).astype(np.int32)

            # the official implementation uses actual indices during the
            # sample_cond_vector method but uses the shuffled version in
            # sampling data, so we're gonna do just that.
            shuffled_idx = np.arange(self.batch_size)
            self._np_rng.shuffle(shuffled_idx)

            shuffled_cat_features_idx = selected_cat_features_idx[shuffled_idx]
            shuffled_values_idx = selected_cat_values_idx[shuffled_idx]

            sample_idx = []
            for feat_id, val_id in zip(shuffled_cat_features_idx,
                                       shuffled_values_idx):
                s_id = self._np_rng.choice(
                    self.row_idx_by_categories[np.squeeze(feat_id)][
                        np.squeeze(val_id)])
                sample_idx.append(np.squeeze(s_id))

            yield x_transformed[sample_idx]



import keras
from keras import random, ops
from keras.backend import backend
from teras._src.utils import clean_reloaded_config_data


class BaseCTGAN(keras.Model):
    """
    Base CTGAN model class.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 metadata: dict,
                 latent_dim: int = 128,
                 seed: int = 1337,
                 **kwargs):
        if not backend() == "jax":
            # Don't call super() with JAX backend as `JAXGAN` does that
            # already!
            super().__init__(**kwargs)
        self.generator = generator
        self.discriminator = discriminator
        self.metadata = metadata
        self.latent_dim = latent_dim
        self.seed = seed
        self._seed_gen = random.SeedGenerator(self.seed)

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(
            name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(
            name="discriminator_loss")

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(
                    learning_rate=1e-3,
                    beta_1=0.5, beta_2=0.9,
                    name="generator_optimizer"
                ),
                discriminator_optimizer=keras.optimizers.Adam(
                    learning_rate=1e-3,
                    beta_1=0.5, beta_2=0.9,
                    name="discriminator_optimizer"
                ),
                **kwargs
                ):
        super().compile(**kwargs)
        self.generator_optimizer = generator_optimizer
        self.discriminator_optimizer = discriminator_optimizer

    def build(self, input_shape):
        batch_size, input_dim = input_shape
        total_cats = self.metadata["categorical"]["total_num_categories"]
        # Generator receives the input of dimensons = data_dim + |cond_vector|
        # where, |cond_vector| = total_num_categories
        input_shape = (batch_size, input_dim + total_cats)
        self.discriminator.build(input_shape)
        # Generator receives the input of dimensons = latent_dim + |cond_vector|
        # where, |cond_vector| = total_num_categories
        input_shape = (batch_size, self.latent_dim + total_cats)
        self.generator.build(input_shape)

    @property
    def metrics(self):
        return [self.generator_loss_tracker,
                self.discriminator_loss_tracker]

    @staticmethod
    def compute_generator_loss(x_generated,
                               y_pred_generated,
                               cond_vectors,
                               mask,
                               metadata):
        """
        Loss for the Generator model in the CTGAN architecture.

        Args:
            x_generated: Samples drawn from the input dataset
            y_pred_generated: Discriminator's output for the generated samples
            cond_vectors: Conditional vectors that are used for and with
                generated samples
            mask: Mask created during the conditional vectors generation step
            metadata: dict, metadata computed during the data transformation step.

        Returns:
            Generator's loss.
        """
        loss = []
        cross_entropy_loss = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=None)
        continuous_features_relative_indices = metadata["continuous"]["relative_indices_all"]
        features_relative_indices_all = metadata["relative_indices_all"]
        num_categories_all = metadata["categorical"]["num_categories_all"]
        # the first k features in the data are continuous which we'll ignore as
        # we're only concerned with the categorical features here
        offset = len(continuous_features_relative_indices)
        for i, index in enumerate(features_relative_indices_all[offset:]):
            logits = x_generated[:, index: index + num_categories_all[i]]
            temp_cond_vector = cond_vectors[:, i: i + num_categories_all[i]]
            labels = ops.argmax(temp_cond_vector, axis=1)
            ce_loss = cross_entropy_loss(y_pred=logits,
                                         y_true=labels
                                         )
            loss.append(ce_loss)
        loss = ops.stack(loss, axis=1)
        loss = ops.sum(loss * ops.cast(mask, dtype="float32")
                       ) / ops.cast(ops.shape(y_pred_generated)[0], dtype="float32")
        loss = -ops.mean(y_pred_generated) * loss
        return loss

    @staticmethod
    def compute_disciriminator_loss(y_pred_real, y_pred_generated):
        """
        Loss for the Discriminator model in the CTGAN architecture.

        Args:
            y_pred_real: Discriminator's output for real samples
            y_pred_generated: Discriminator's output for generated samples

        Returns:
            Discriminator's loss.
        """
        return -(ops.mean(y_pred_real) - ops.mean(y_pred_generated))

    def test_step(self, data):
        x, cond_vectors_real, cond_vectors, mask = data
        batch_size = ops.shape(x)[0]

        # =========================
        # Discriminator test step
        # =========================
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=self._seed_gen)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)
        x_generated = self.generator(input_gen)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        x = ops.concatenate([x, cond_vectors_real],
                            axis=1)

        y_pred_generated = self.discriminator(x_generated)
        y_pred_real = self.discriminator(x)
        loss_disc = self.compute_disciriminator_loss(
            y_pred_real,
            y_pred_generated)

        # =====================
        # Generator test step
        # =====================
        z = random.normal(shape=[batch_size, self.latent_dim],
                          seed=self._seed_gen)
        input_gen = ops.concatenate([z, cond_vectors], axis=1)

        x_generated = self.generator(input_gen)
        x_generated = ops.concatenate(
            [x_generated, cond_vectors], axis=1)
        y_pred_generated = self.discriminator(x_generated)
        loss_gen = self.compute_generator_loss(
            x_generated, y_pred_generated,
            cond_vectors=cond_vectors, mask=mask,
            metadata=self.metadata)

        self.generator_loss_tracker.update_state(loss_gen)
        self.discriminator_loss_tracker.update_state(loss_disc)
        results = {m.name: m.result() for m in self.metrics}
        return results

    def predict_step(self, generator_input):
        generated_samples = self.generator(generator_input)
        return generated_samples

    def get_config(self):
        config = super().get_config()
        config.update({
            'generator': keras.layers.serialize(self.generator),
            'discriminator': keras.layers.serialize(self.generator),
            'latent_dim': self.latent_dim,
            'seed': self.seed,
        }
        )
        return config

    @classmethod
    def from_config(cls, config):
        generator = keras.layers.deserialize(config.pop("generator"))
        discriminator = keras.layers.deserialize(config.pop("discriminator"))
        metadata = clean_reloaded_config_data(config.pop("metadata"))
        return cls(generator=generator,
                   discriminator=discriminator,
                   metadata=metadata,
                   **config)



# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from teras._src.typing import (
    Number as Number,
    ActivationType as ActivationType,
    IntegerSequence as IntegerSequence,
    LayersList as LayersList,
    LayersCollection as LayersCollection,
    NormalizationType as NormalizationType,
    FloatSequence as FloatSequence,
    InitializerType as InitializerType,
    LayerOrModelType as LayerOrModelType,
    FeaturesNamesType as FeaturesNamesType,
    OptionalIntOrStr as OptionalIntOrStr,
    ListOrArray as ListOrArray,
    DataFrameOrNdArray as DataFrameOrNdArray
)


import keras

from teras._src.models.tasks.task import Task
from teras._src.typing import ActivationType
from teras._src.api_export import teras_export


@teras_export("teras.models.Classifier")
class Classifier(Task):
    """
    Classifier class that provides a dense prediction head.

    Args:
        backbone: `keras.Model` instance, backbone is called on
            inputs followed by the dense head that produces predictions.
        num_classes: int, number of classes to predict.
        activation: str or callable, activation function to use for
            outputs. Defaults to "softmax"
        hidden_dim: int, hidden dimensionality of the dense head.
            Defaults to 1024.
        hidden_activation: str or callable, activation for the hidden layer.
            Defaults to "relu".
    """
    def __init__(self,
                 backbone: keras.Model,
                 num_classes: int,
                 activation: ActivationType = "softmax",
                 hidden_dim: int = 1024,
                 hidden_activation: ActivationType = "relu",
                 **kwargs):
        inputs = backbone.input
        x = backbone(inputs)
        # In case the backbone outputs are of shape (None, a, b),
        # for instance in the case of transformer based models
        x = keras.layers.Flatten()(x)
        x = keras.layers.Dense(hidden_dim, activation=hidden_activation,
                               name="hidden_layer_classification_head")(x)
        outputs = keras.layers.Dense(num_classes,
                                     activation=activation,
                                     name="predictions")(x)
        super().__init__(inputs, outputs, **kwargs)

        self.backbone = backbone
        self.num_classes = num_classes
        self.activation = activation
        self.hidden_dim = hidden_dim
        self.hidden_activation = hidden_activation

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "backbone": keras.layers.serialize(self.backbone),
                "num_classes": self.num_classes,
                "activation": self.activation,
                "hidden_dim": self.hidden_dim,
                "hidden_activation": self.hidden_activation
            }
        )



from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.utils.create_gain_dataset")
def create_gain_dataset(x, seed: int = 1337):
    """
    Creates a torch dataloader compatible with the `GAIN` architecture.
    The resultant dataset produces a tuple consisting of a batch of data for
    `generator` and another batch of data for the `discriminator`.

    Args:
        x: Dataset to use for training.
        seed: int, seed to use in shuffling. Defaults to 1337.
    """
    return backend.utils.create_gain_dataset(x, seed)




import time
import jax
import numpy as np
import keras
from keras import random, ops
from keras.backend import floatx
from teras._src.backend.generic_utils import dataset_type

# If it works, it works.

# Define variables for GAIN
_generator: keras.Model = None
_discriminator: keras.Model = None
_hint_rate = None
_alpha = None
_seed = None
_generator_optimizer = None
_discriminator_optimizer = None
_models_built = False
_generator_loss_tracker = keras.metrics.Mean(name="generator_loss")
_discriminator_loss_tracker = keras.metrics.Mean(name="discriminator_loss")

_initialized: bool = False
_compiled: bool = False


def init(generator: keras.Model, discriminator: keras.Model,
         hint_rate: float = 0.9, alpha: float = 100., seed: int = 1337):
    """
    `init` method analogous to `keras.Model.__init__()`.


    Args:
        generator: keras.Model, instance of `GAINGenerator`
        discriminator: keras.Model, instance of `GAINDiscrimniator`.
        hint_rate: float, Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the underlying
            data distribution.
            Defaults to 0.9
        alpha: float, Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely,
            `generator_loss` = `cross_entropy_loss` + `alpha` * `mse_loss`
            The higher the `alpha`, the more the mse_loss will affect the
            overall generator loss.
            Defaults to 100.
        seed: int, seed for random ops.
    """
    global _generator
    global _discriminator
    global _hint_rate
    global _alpha
    global _seed
    global _initialized
    _generator = generator
    _discriminator = discriminator
    _hint_rate = hint_rate
    _alpha = alpha
    _seed = seed

    _initialized = True


def compile(generator_optimizer=keras.optimizers.Adam(
                name="generator_optimizer"),
            discriminator_optimizer=keras.optimizers.Adam(
                name="discriminator_optimizer")):
    """
    Compile method analogous to `keras.Model.compile()`.

    Args:
        generator_optimizer: Optimizer for generator.
        discriminator_optimizer: Optimizer for discriminator.
    """
    global _generator_optimizer
    global _discriminator_optimizer
    global _compiled

    _generator_optimizer = generator_optimizer
    _discriminator_optimizer = discriminator_optimizer

    _compiled = True


def build(generator_input_shape, discriminator_input_shape):
    """
    `build` method analogous to `keras.Model.build()`.

    Args:
        generator_input_shape: Expected input shape for generator.
        discriminator_input_shape: Expected input shape for discriminator.
    """
    global _models_built
    global _generator
    global _discriminator
    _generator.build(generator_input_shape)
    _discriminator.build(discriminator_input_shape)

    _models_built = True


def metrics():
    """
    Returns:
        List of metrics.
    """
    return [_generator_loss_tracker,
            _discriminator_loss_tracker]


def fit(x, epochs: int = 1, verbose: bool = True):
    """
    Trainer function for GAIN for JAX backend.

    Args:
        x: Dataset. Create it using `teras.utils.create_gain_dataset` function.
        epochs: int, Number of epochs to train.
        verbose: bool, whether to print training logs to console.

    Returns:
        A tuple of trained `generator_state` and `discriminator_state`.
    """
    global _models_built
    global _initialized
    global _compiled
    global _hint_rate
    global _alpha
    global _seed
    if not _initialized:
        raise RuntimeError(
            "Trainer not initialized. Please call `Trainer.init` method "
            "before calling `fit`. "
        )

    if not _compiled:
        raise RuntimeError(
            "Trainer not compiled. Please call `Trainer.compile()` method "
            "before calling `fit`. "
        )

    _optimizers_built = False
    total_batches = 0
    if dataset_type(x) == "not_supported":
        raise ValueError(
            "Unsupported type for `x`. "
            "It should be tensorflow dataset, pytorch dataloader."
            f"But received {type(x)}")
    aux_args = (_hint_rate, _alpha)
    key = jax.random.PRNGKey(_seed)
    is_first_iteration = True
    for epoch in range(epochs):
        epoch_start_time = time.time()
        for batch_num, batch in enumerate(x):
            batch = np.asarray(batch)
            if is_first_iteration:
                if not _models_built:
                    if isinstance(batch, tuple):
                        input_shape = ops.shape(batch[0])
                    else:
                        input_shape = ops.shape(batch)
                    input_shape = (input_shape[:-1], input_shape[-1] * 2)
                    _generator.build(input_shape)
                    _discriminator.build(input_shape)
                    _models_built = True

                if not _optimizers_built:
                    _generator_optimizer.build(_generator.trainable_variables)
                    _discriminator_optimizer.build(
                        _discriminator.trainable_variables)
                    _optimizers_built = True

                # Create initial state
                generator_state = (
                    _generator.trainable_variables,
                    _generator.non_trainable_variables,
                    _generator_optimizer.variables
                )

                discriminator_state = (
                    _discriminator.trainable_variables,
                    _discriminator.non_trainable_variables,
                    _discriminator_optimizer.variables
                )

                is_first_iteration = False

            logs, generator_state, discriminator_state = train_step(
                generator_state,
                discriminator_state,
                batch,
                aux_args,
                jax.random.split(key, 1)[0]
            )
            if verbose:
                epoch_str = f"Epoch {epoch + 1}/{epochs}"
                elapsed_time_str = (f"Elapsed "
                                    f"{time.time() - epoch_start_time:.2f}s")
                if total_batches > 0 and epoch > 0:
                    batch_str = f"Batch {batch_num + 1}/{total_batches}"
                else:
                    batch_str = f"Batch {batch_num + 1}/?"
                    total_batches += 1
                logs_str = (f"generator_loss: "
                            f"{logs['generator_loss']:.4f}   "
                            f"discriminator_loss: "
                            f"{logs['discriminator_loss']:.4f}   "
                            )
                print(
                    f"\r{epoch_str:<15} {elapsed_time_str:<15} {batch_str:<15}"
                    f" {logs_str}",
                    end="")
        print()
        _trained = True

    return generator_state, discriminator_state


def compute_discriminator_loss(mask, mask_pred):
    return keras.losses.BinaryCrossentropy()(mask, mask_pred)


def compute_generator_loss(x, x_generated, mask, mask_pred, alpha):
    cross_entropy_loss = keras.losses.CategoricalCrossentropy()(
        mask, mask_pred
    )
    mse_loss = keras.losses.MeanSquaredError()(
        y_true=(mask * x),
        y_pred=(mask * x_generated))
    loss = cross_entropy_loss + alpha * mse_loss
    return loss


def generator_compute_loss_and_updates(
        generator_trainable_vars,
        generator_non_trainable_vars,
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_gen,
        hint_vectors,
        mask,
        alpha,
        training=False
):
    x_generated, generator_non_trainable_vars = _generator.stateless_call(
        generator_trainable_vars,
        generator_non_trainable_vars,
        ops.concatenate([x_gen, mask], axis=1),
        training=training,
    )
    x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
    mask_pred, _ = (
        _discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            ops.concatenate([x_hat, hint_vectors], axis=1),
            training=training,
        )
    )
    loss = compute_generator_loss(
        x=x_gen,
        x_generated=x_generated,
        mask=mask,
        mask_pred=mask_pred,
        alpha=alpha
    )
    return loss, (generator_non_trainable_vars,)


def discriminator_compute_loss_and_updates(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_hat_disc,
        hint_vectors,
        mask,
        training=False
):
    mask_pred, discriminator_non_trainable_vars = (
        _discriminator.stateless_call(
            discriminator_trainable_vars,
            discriminator_non_trainable_vars,
            ops.concatenate([x_hat_disc, hint_vectors], axis=1),
            training=training,
        )
    )
    loss = compute_discriminator_loss(mask, mask_pred)
    return loss, (discriminator_non_trainable_vars,)


@jax.jit
def train_step(generator_state, discriminator_state, data, aux_args, key):
    (
        generator_trainable_vars,
        generator_non_trainable_vars,
        generator_optimizer_vars,
    ) = generator_state

    (
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        discriminator_optimizer_vars,
    ) = discriminator_state

    (
        hint_rate,
        alpha,
    ) = aux_args

    # data is a tuple of x_generator and x_discriminator batches
    # drawn from the dataset. The reason behind generating two separate
    # batches of data at each step is that it's how GAIN's algorithm works
    x_gen, x_disc = data

    # =========================
    # Train the discriminator
    # =========================
    # Create mask
    mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
    # replace nans with 0.
    x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
    # Sample noise
    z = random.uniform(shape=ops.shape(x_disc), minval=0., maxval=0.01,
                       seed=key)
    # Sample hint vectors
    hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                   counts=1,
                                   probabilities=hint_rate,
                                   seed=key)
    hint_vectors = hint_vectors * mask
    # Combine random vectors with original data
    x_disc = x_disc * mask + (1 - mask) * z

    x_generated, _ = _generator.stateless_call(
        generator_trainable_vars,
        generator_non_trainable_vars,
        ops.concatenate([x_disc, mask], axis=1),
        training=False,
    )
    x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)

    disc_grad_fn = jax.value_and_grad(
        discriminator_compute_loss_and_updates,
        has_aux=True,
    )
    (d_loss, (discriminator_non_trainable_vars,)), grads = disc_grad_fn(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_hat_disc,
        hint_vectors,
        mask,
        training=True,
    )

    (
        discriminator_trainable_vars,
        discriminator_optimizer_vars
    ) = _discriminator_optimizer.stateless_apply(
        discriminator_optimizer_vars,
        grads,
        discriminator_trainable_vars
    )

    discriminator_state = (
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        discriminator_optimizer_vars
    )

    # =====================
    # Train the generator
    # =====================
    mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
    x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
    z = random.uniform(shape=ops.shape(x_gen), minval=0., maxval=0.01,
                       seed=key)
    hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                   counts=1,
                                   probabilities=hint_rate,
                                   seed=key)
    hint_vectors = hint_vectors * mask
    x_gen = x_gen * mask + (1 - mask) * z

    gen_grad_fn = jax.value_and_grad(
        generator_compute_loss_and_updates,
        has_aux=True,
    )
    (g_loss, (generator_non_trainable_vars,)), grads = gen_grad_fn(
        generator_trainable_vars,
        generator_non_trainable_vars,
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_gen,
        hint_vectors,
        mask,
        alpha,
        training=True,
    )

    (
        generator_trainable_vars,
        generator_optimizer_vars
    ) = _generator_optimizer.stateless_apply(
        generator_optimizer_vars,
        grads,
        generator_trainable_vars
    )

    generator_state = (
        generator_trainable_vars,
        generator_non_trainable_vars,
        generator_optimizer_vars
    )

    # Update custom tracking metrics
    _generator_loss_tracker.update_state(g_loss)
    _discriminator_loss_tracker.update_state(d_loss)

    logs = {m.name: m.result() for m in metrics()}
    return logs, generator_state, discriminator_state



import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.backbones.tab_transformer.tab_transformer import \
    TabTransformerBackbone
from teras._src.models.pretrainers.tab_transformer import (
    TabTransformerMLMPretrainer,
    TabTransformerRTDPretrainer
)


class TabTransformerMLMPretrainerTest(TestCase):
    def setUp(self):
        # we assume there are 4 categorical features in the dataset,
        # with feature 1 having 5 unique categories, feature 2 having 3
        # unique categories, feature 3 with 2 unique categories and
        # feature 4 with 7 unique categories
        self.batch_size = 16
        x_cat = ops.array([
            random.randint((self.batch_size,), minval=0, maxval=card)
            for card in [5, 3, 2, 7]],
            dtype="float32")
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((self.batch_size, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous
        self.num_categorical = 4    # Number of categorical features
        self.num_continuous = 3     # Number of continuous features
        self.embedding_dim = 16
        self.mask = random.binomial((self.batch_size, 7),
                                    1, 0.5)

    def test_valid_call(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerMLMPretrainer(
            model=tab_model,
            data_dim=7
        )
        reconstructed_features = pretrainer(self.input_batch,
                                            self.mask)

    def test_valid_output_shape(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerMLMPretrainer(
            model=tab_model,
            data_dim=7
        )
        reconstructed_features = pretrainer(self.input_batch,
                                            self.mask)
        self.assertEqual(ops.shape(reconstructed_features),
                         (self.batch_size, 7))

    def test_fit(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerMLMPretrainer(
            model=tab_model,
            data_dim=7
        )
        # pretrainer.compile(loss=keras.losses.CategoricalCrossentropy(
        #     from_logits=True),
        #     optimizer=(keras.optimizers.Adam()))
        pretrainer.compile()
        pretrainer.build(ops.shape(self.input_batch))
        pretrainer.fit(self.input_batch)


class TabTransformerRTDPretrainerTest(TestCase):
    def setUp(self):
        # we assume there are 4 categorical features in the dataset,
        # with feature 1 having 5 unique categories, feature 2 having 3
        # unique categories, feature 3 with 2 unique categories and
        # feature 4 with 7 unique categories
        self.batch_size = 16
        x_cat = ops.array([
            random.randint((self.batch_size,), minval=0, maxval=card)
            for card in [5, 3, 2, 7]],
            dtype="float32")
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((self.batch_size, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous
        self.num_categorical = 4    # Number of categorical features
        self.num_continuous = 3     # Number of continuous features
        self.embedding_dim = 16
        self.mask = random.binomial((self.batch_size, 7),
                                    1, 0.5)

    def test_valid_call(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerRTDPretrainer(
            model=tab_model,
            data_dim=7
        )
        predicted_mask = pretrainer(self.input_batch,
                                    self.mask)

    def test_valid_output_shape(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerRTDPretrainer(
            model=tab_model,
            data_dim=7
        )
        predicted_mask = pretrainer(self.input_batch,
                                    self.mask)
        self.assertEqual(ops.shape(predicted_mask),
                         (self.batch_size, 7))

    def test_fit(self):
        tab_model = TabTransformerBackbone(
            input_dim=7,
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        pretrainer = TabTransformerRTDPretrainer(
            model=tab_model,
            data_dim=7
        )
        pretrainer.compile(optimizer=keras.optimizers.Adam())
        pretrainer.build(ops.shape(self.input_batch))
        pretrainer.fit(self.input_batch)



# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from teras._src.backend.tensorflow.models.pretrainers.tabnet import TabNetPretrainer
from teras._src.backend.tensorflow.models.pretrainers.tab_transformer import (
    TabTransformerMLMPretrainer,
    TabTransformerRTDPretrainer
)
from teras._src.backend.tensorflow.models.pretrainers.saint import SAINTPretrainer

# GANs
from teras._src.backend.tensorflow.models.gans.gain import GAIN
from teras._src.backend.tensorflow.models.gans.pcgain import PCGAIN
from teras._src.backend.tensorflow.models.gans.ctgan.discriminator import CTGANDiscriminator
from teras._src.backend.tensorflow.models.gans.ctgan.generator import CTGANGenerator
from teras._src.backend.tensorflow.models.gans.ctgan.ctgan import CTGAN

# Autoencoders
from teras._src.backend.tensorflow.models.autoencoders.tvae.tvae import TVAE



import keras
from keras import random, ops
from teras._src.backend.common.models.gans.gain import BaseGAIN
from keras.backend import floatx


class BasePCGAIN(BaseGAIN):
    """
    Base class for PCGAIN.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 classifier: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 200.,
                 beta: float = 100.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         seed=seed,
                         **kwargs)
        self.classifier = classifier
        self.beta = beta

    def build(self, input_shape):
        super().build(input_shape)
        # we want to build classifier last so its variables are at the end of
        # lists for both trainable and non-trainable variables lists
        self.classifier.build(input_shape)

    def compute_generator_loss(self, x, x_generated, mask, mask_pred,
                               classifier_pred, alpha, beta):
        cross_entropy_loss = keras.losses.BinaryCrossentropy()(
            mask, mask_pred
        )
        mse_loss = keras.losses.MeanSquaredError()(
            y_true=(mask * x),
            y_pred=(mask * x_generated))

        info_entropy_loss = -ops.mean(
            classifier_pred * ops.log(classifier_pred + 1e-8))
        loss = cross_entropy_loss + (alpha * mse_loss) + (beta *
                                                          info_entropy_loss)
        return loss

    def test_step(self, data):
        x_gen, x_disc = data

        # =========================
        # Test the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        z = random.uniform(shape=ops.shape(x_disc), minval=0., maxval=0.01,
                           seed=self._seed_gen)
        # Sample hint vectors
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z
        x_generated = self.generator(ops.concatenate([x_disc, mask], axis=1))
        # Combine generated samples with original data
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat_disc, hint_vectors], axis=1))
        d_loss = self.compute_discriminator_loss(mask, mask_pred)

        # =====================
        # Test the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        z = random.uniform(shape=ops.shape(x_gen), minval=0., maxval=0.01,
                           seed=self._seed_gen)
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z

        x_generated = self.generator(
            ops.concatenate([x_gen, mask], axis=1))
        # Combine generated samples with original/observed data
        x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
        classifier_pred = self.classifier(x_hat)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat, hint_vectors], axis=1))
        g_loss = self.compute_generator_loss(
            x=x_gen,
            x_generated=x_generated,
            mask=mask,
            mask_pred=mask_pred,
            classifier_pred=classifier_pred,
            alpha=self.alpha,
            beta=self.beta,
        )

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(g_loss)
        self.discriminator_loss_tracker.update_state(d_loss)

        logs = {m.name: m.result() for m in self.metrics}
        return logs

    def predict_step(self, data):
        """
        Args:
            Transformed data.
        Returns:
            Imputed data that should be reverse transformed
            to its original form.
        """
        if isinstance(data, tuple):
            data = data[0]
        data = ops.cast(data, floatx())
        # Create mask
        mask = 1. - ops.cast(ops.isnan(data), dtype=floatx())
        data = ops.where(ops.isnan(data), x1=0., x2=data)
        # Sample noise
        z = random.uniform(ops.shape(data), minval=0., maxval=0.01,
                           seed=self._seed_gen)
        x = mask * data + (1 - mask) * z
        imputed_data = self.generator(ops.concatenate([x, mask], axis=1))
        imputed_data = mask * data + (1 - mask) * imputed_data
        return imputed_data

    def get_config(self):
        config = super().get_config()
        config.update({
            'classifier': keras.layers.serialize(self.classifier),
            'beta': keras.layers.serialize(self.beta),
        })
        return config

    @classmethod
    def from_config(cls, config):
        generator = keras.layers.deserialize(config.pop("generator"))
        discriminator = keras.layers.deserialize(config.pop("discriminator"))
        classifier = keras.layers.deserialize(config.pop("classifier"))
        return cls(generator=generator, discriminator=discriminator,
                   classifier=classifier, **config)



# This module contains wrapper functions around some Keras 3 functions
# to make them more easy and intuitive to use.
import keras
from keras import ops, random


def binomial(shape: tuple):
    """
    Samples from a binomial distribution of a given shape
    Args:
        shape: 

    Returns:

    """
    pass


import os

import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.backbones.tab_transformer.tab_transformer import \
    TabTransformerBackbone


class TabTransformerBackboneTest(TestCase):
    def setUp(self):
        # we assume there are 4 categorical features in the dataset,
        # with feature 1 having 5 unique categories, feature 2 having 3
        # unique categories, feature 3 with 2 unique categories and
        # feature 4 with 7 unique categories
        x_cat = ops.array([
            random.randint((16,), minval=0, maxval=card)
            for card in [5, 3, 2, 7]],
            dtype="float32")
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((16, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous
        self.num_categorical = 4 # Number of categorical features
        self.num_continuous = 3 # Number of continuous features
        self.embedding_dim = 16

    def test_valid_call(self):
        model = TabTransformerBackbone(
            input_dim=self.input_batch.shape[1],
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        outputs = model(self.input_batch)

    def test_valid_output_shape(self):
        model = TabTransformerBackbone(
            input_dim=self.input_batch.shape[1],
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        outputs = model(self.input_batch)
        # Outputs must have the following shape,
        # (batch_size, `num_categorical` * `embedding_dim` + `num_continuous)
        true_output_shape = (
            self.input_batch.shape[0],
            self.num_categorical * self.embedding_dim + self.num_continuous
        )
        self.assertEqual(true_output_shape, ops.shape(outputs))

    def test_model_save_and_load(self):
        model = TabTransformerBackbone(
            input_dim=self.input_batch.shape[1],
            cardinalities=self.cardinalities,
            embedding_dim=self.embedding_dim
        )
        outputs = model(self.input_batch)
        save_path = os.path.join(self.get_temp_dir(),
                                 "tab_transformer_backbone.keras"
                                 )
        model.save(save_path)
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TabTransformerBackbone)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.input_batch)
        self.assertAllClose(
            ops.convert_to_numpy(outputs),
            ops.convert_to_numpy(reloaded_outputs)
        )



# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Activation
from teras._src.layers.activation.gumbel_softmax import (
    GumbelSoftmax as GumbelSoftmax)

# CTGAN
from teras._src.layers.ctgan.generator_layer import (
    CTGANGeneratorLayer as CTGANGeneratorLayer)
from teras._src.layers.ctgan.discriminator_layer import (
    CTGANDiscriminatorLayer as CTGANDiscriminatorLayer)

# FT-Transformer
from teras._src.layers.ft_transformer.feature_tokenizer import (
    FTTransformerFeatureTokenizer as FTTransformerFeatureTokenizer)

# SAINT
from teras._src.layers.saint.embedding import (
    SAINTEmbedding as SAINTEmbedding)
from teras._src.layers.saint.encoder_layer import (
    SAINTEncoderLayer as SAINTEncoderLayer)
from teras._src.layers.saint.multi_head_inter_sample_attention import (
    SAINTMultiHeadInterSampleAttention as SAINTMultiHeadInterSampleAttention)
from teras._src.layers.saint.projection_head import (
    SAINTProjectionHead as SAINTProjectionHead)
from teras._src.layers.saint.reconstruction_head import (
    SAINTReconstructionHead as SAINTReconstructionHead)

# TabTransformer
from teras._src.layers.tab_transformer.column_embedding import (
    TabTransformerColumnEmbedding as TabTransformerColumnEmbedding
)

# TabNet
from teras._src.layers.tabnet.attentive_transformer import (
    TabNetAttentiveTransformer as TabNetAttentiveTransformer)
from teras._src.layers.tabnet.feature_transformer_layer import (
    TabNetFeatureTransformerLayer as TabNetFeatureTransformerLayer)
from teras._src.layers.tabnet.feature_transformer import (
    TabNetFeatureTransformer as TabNetFeatureTransformer)

# Transformer
from teras._src.layers.transformer.encoder_layer import (
    TransformerEncoderLayer as TransformerEncoderLayer)
from teras._src.layers.transformer.feedforward import (
    TransformerFeedForward as TransformerFeedForward)

# Categorical Embedding
from teras._src.layers.categorical_embedding import (
    CategoricalEmbedding as CategoricalEmbedding)

# Categorical Extraction
from teras._src.layers.categorical_extraction import (
    CategoricalExtraction as CategoricalExtraction)

# CLS Token related
from teras._src.layers.cls_token import CLSToken as CLSToken
from teras._src.layers.cls_token_extraction import (
    CLSTokenExtraction as CLSTokenExtraction)

# Continous Extraction
from teras._src.layers.continuous_extraction import (
    ContinuousExtraction as ContinuousExtraction)

# CutMix
from teras._src.layers.cutmix import CutMix as CutMix

# MixUp
from teras._src.layers.mixup import MixUp as MixUp

# LayerList
from teras._src.layers.layer_list import LayerList as LayerList



import jax
from jax import lax
import keras
from keras import ops
from teras._src.backend.common.models.autoencoders.tvae.tvae import BaseTVAE


class TVAE(BaseTVAE):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 metadata: dict,
                 data_dim: int,
                 latent_dim: int = 128,
                 loss_factor: float = 2.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(encoder=encoder,
                         decoder=decoder,
                         metadata=metadata,
                         data_dim=data_dim,
                         latent_dim=latent_dim,
                         loss_factor=loss_factor,
                         seed=seed,
                         **kwargs)

    def compute_loss(self, real_samples=None, generated_samples=None,
                     sigmas=None, mean=None, log_var=None, metadata=None,
                     data_dim=None, loss_factor=None):
        loss = []
        cross_entropy = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True,
            reduction="sum")
        cont_i = 0  # continuous index
        cat_i = 0  # categorical index
        for i, relative_index in enumerate(metadata["relative_indices_all"]):
            # the first k features are continuous
            if i < metadata["num_continuous"]:
                # each continuous feature is of the form
                # [alpha, beta1, beta2...beta(n)] where n is the number of
                # clusters

                # calculate alpha loss
                std = sigmas[relative_index]
                eq = real_samples[:, relative_index] - ops.tanh(
                    generated_samples[:, relative_index])
                loss_temp = ops.sum((eq ** 2 / 2 / (std ** 2)))
                loss.append(loss_temp)
                loss_temp = ops.log(std) * ops.cast(
                    ops.shape(real_samples)[0], dtype=std.dtype)
                loss.append(loss_temp)

                # calculate betas loss
                num_clusters = metadata["continuous"][
                    "num_valid_clusters_all"][cont_i]
                logits = lax.dynamic_slice_in_dim(
                    generated_samples,
                    start_index=relative_index + 1,
                    slice_size=num_clusters
                )
                labels = ops.argmax(
                    lax.dynamic_slice_in_dim(
                        real_samples,
                        start_index=relative_index + 1,
                        slice_size=num_clusters,
                    ),
                    axis=-1
                )
                cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
                loss.append(cross_entropy_loss)
                cont_i += 1
            else:
                num_categories = metadata["categorical"][
                    "num_categories_all"][cat_i]
                logits = lax.dynamic_slice_in_dim(
                    generated_samples,
                    start_index=relative_index,
                    slice_size=num_categories
                )
                labels = ops.argmax(
                    lax.dynamic_slice_in_dim(
                        real_samples,
                        start_index=relative_index,
                        slice_size=num_categories
                    ),
                    axis=-1
                )
                logits = ops.squeeze(logits)
                cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
                loss.append(cross_entropy_loss)
                cat_i += 1
        KLD = -0.5 * ops.sum(1 + log_var - mean ** 2 - ops.exp(log_var))
        loss_1 = ops.sum(loss) * loss_factor / data_dim
        loss_2 = KLD / data_dim
        final_loss = loss_1 + loss_2
        return final_loss

    def compute_loss_and_updates(self, trainable_variables,
                                 non_trainable_variables, x):
        (
            (generated_samples, sigmas, mean, log_var),
            non_trainable_variables
        ) = self.stateless_call(
            trainable_variables,
            non_trainable_variables,
            x,
        )
        loss = self.compute_loss(real_samples=x,
                                 generated_samples=generated_samples,
                                 sigmas=sigmas, mean=mean, log_var=log_var,
                                 metadata=self.metadata,
                                 data_dim=self.data_dim,
                                 loss_factor=self.loss_factor)
        return loss, (non_trainable_variables, sigmas)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state
        grad_fn = jax.value_and_grad(self.compute_loss_and_updates,
                                     has_aux=True)
        (loss, (non_trainable_variables, sigmas)), grads = grad_fn(
            trainable_variables,
            non_trainable_variables,
            data
        )
        (
            trainable_variables,
            optimizer_variables
        ) = self.optimizer.stateless_apply(optimizer_variables,
                                           grads,
                                           trainable_variables)
        sigmas = ops.clip(sigmas, x_min=0.01, x_max=1.0)
        self.decoder.sigmas = sigmas
        # since we only have one metric i.e. loss.
        metrics_variables = self.loss_tracker.stateless_update_state(
            metrics_variables,
            loss)
        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        )
        logs = {"loss": self.loss_tracker.stateless_result(metrics_variables)}
        return logs, state

    def test_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        ) = state
        (loss,
         (non_trainable_variables, sigmas)
         ), grads = self.compute_loss_and_updates(
            trainable_variables,
            non_trainable_variables,
            data
        )
        # since we only have one metric i.e. loss.
        metrics_variables = self.loss_tracker.stateless_update_state(
            metrics_variables,
            loss)
        state = (
            trainable_variables,
            non_trainable_variables,
            metrics_variables
        )
        logs = {"loss": self.loss_tracker.stateless_result(metrics_variables)}
        return logs, state

    def predict_step(self, state, z):
        trainable_variables, non_trainable_variables = state
        # Since encoder comes hence gets built before decoder, so the parent
        # model's list of vars contains encoder vars followed by decoder vars.
        # if you didn't know that, feel free to thank me! jk. happy coding!
        start_idx = len(self.encoder.trainable_variables)
        slice_size = len(self.decoder.trainable_variables)
        decoder_trainable_vars = lax.dynamic_slice_in_dim(
            trainable_variables,
            start_index=start_idx,
            slice_size=slice_size
        )
        start_idx = len(self.encoder.non_trainable_variables)
        slice_size = len(self.decoder.non_trainable_variables)
        decoder_non_trainable_vars = lax.dynamic_slice_in_dim(
            non_trainable_variables,
            start_index=start_idx,
            slice_size=slice_size
        )
        (
            (generated_samples, sigmas),
            decoder_non_trainable_vars
        ) = self.decoder.stateless_call(
            decoder_trainable_vars,
            decoder_non_trainable_vars,
            z
        )
        return generated_samples, decoder_non_trainable_vars



import keras
import numpy as np
from keras import backend
from keras.src.testing.test_case import TestCase

from teras._src.models.gans.gain.discriminator import GAINDiscriminator
from teras._src.models.gans.gain.generator import GAINGenerator
from teras._src.models.gans.pcgain.classifier import PCGAINClassifier
from teras._src.models.gans.pcgain.pcgain import PCGAIN
from teras._src.data_utils import create_gain_dataset
from teras._src.utils import inject_missing_values


class PCGAINTest(TestCase):
    def setUp(self):
        self.data_dim = 7
        self.rng = np.random.default_rng(1337)
        x_gen = self.rng.uniform(0., high=10., size=(8, self.data_dim)
                                 ).astype(np.float32)
        x_gen = inject_missing_values(x_gen)
        x_disc = self.rng.uniform(0., high=10., size=(8, self.data_dim)
                                  ).astype(np.float32)
        x_disc = inject_missing_values(x_disc)
        self.input_ds = create_gain_dataset(x_gen)
        if keras.backend.backend() == "torch":
            from torch.utils.data import DataLoader
            self.input_ds = DataLoader(self.input_ds, batch_size=8)
        else:
            self.input_ds = self.input_ds.batch(batch_size=8)

    def test_fit(self):
        generator = GAINGenerator(data_dim=self.data_dim)
        discriminator = GAINDiscriminator(data_dim=self.data_dim)
        classifier = PCGAINClassifier(num_classes=4,
                                      data_dim=self.data_dim)
        pcgain = PCGAIN(generator,
                        discriminator,
                        classifier)
        pcgain.compile(generator_optimizer=keras.optimizers.Adam(),
                       discriminator_optimizer=keras.optimizers.Adam())
        pcgain.build((8, self.data_dim))
        if backend.backend() == "jax":
            pcgain.build_optimizers()
        logs = pcgain.fit(self.input_ds)

    def test_predict(self):
        generator = GAINGenerator(data_dim=self.data_dim)
        discriminator = GAINDiscriminator(data_dim=self.data_dim)
        classifier = PCGAINClassifier(num_classes=4,
                                      data_dim=self.data_dim)
        pcgain = PCGAIN(generator,
                        discriminator,
                        classifier)
        pcgain.compile(generator_optimizer=keras.optimizers.Adam(),
                       discriminator_optimizer=keras.optimizers.Adam())
        pcgain.build((8, self.data_dim))
        if backend.backend() == "jax":
            pcgain.build_optimizers()
        logs = pcgain.fit(self.input_ds)
        x_test = self.rng.uniform(0., high=10., size=(8, self.data_dim))
        x_test = inject_missing_values(x_test)
        x_imputed = pcgain.predict(x_test)
        self.assertEqual(np.isnan(x_imputed).sum(), 0)



from teras._src.layers.saint.encoder_layer import SAINTEncoderLayer
from keras.src.testing.test_case import TestCase
from keras import ops, random


class SAINTEncoderLayerTest(TestCase):
    def setUp(self):
        self.input_shape = (8, 5, 16)
        self.input_batch = random.normal(self.input_shape)
        self.embedding_dim = self.input_shape[-1]

    def test_valid_output_shape(self):
        encoder = SAINTEncoderLayer(embedding_dim=self.embedding_dim)
        outputs = encoder(self.input_batch)
        self.assertEqual(ops.shape(outputs), self.input_shape)
    


import keras
from teras._src.activations import glu
from teras._src.api_export import teras_export


@teras_export("teras.layers.TabNetFeatureTransformerLayer")
class TabNetFeatureTransformerLayer(keras.layers.Layer):
    """
    TabNetFeatureTransformerLayer layer that serves as the building block
    for the `TabNetFeatureTransformer` layer which is proposed by Arik
    et al. in the "TabNet: Attentive Interpretable Tabular Learning" paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        dim: int, the dense layer first maps the inputs to dim * 2
            dimension hidden representations and later the glu activation
            maps the hidden representations to `dim`-dimensions.
        batch_momentum: float, batch momentum
    """
    def __init__(self,
                 dim: int,
                 batch_momentum: float,
                 **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.batch_momentum = batch_momentum

        self.dense = keras.layers.Dense(self.dim * 2,
                                        use_bias=False)
        self.batch_norm = keras.layers.BatchNormalization(
            momentum=self.batch_momentum)

    def build(self, input_shape):
        self.dense.build(input_shape)
        input_shape = self.dense.compute_output_shape(input_shape)
        self.batch_norm.build(input_shape)

    def call(self, inputs):
        x = self.dense(inputs)
        x = self.batch_norm(x)
        return glu(x)

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "dim": self.dim,
            "batch_momentum": self.batch_momentum,
        })
        return config



from keras import ops, random
from teras._src.api_export import teras_export


# =========================== GLU ===========================
@teras_export("teras.activations.glu")
def glu(logits, axis: int = -1):
    """
    Generalized linear unit nonlinear activation.

    Args:
        logits: `Tensor`, tensor of logits.
        axis: `int`, axis along which to apply glu activation.
            Defaults to -1.
    """
    x, gates = ops.split(logits,
                         indices_or_sections=2,
                         axis=axis)
    return x * ops.sigmoid(gates)


# =========================== GEGLU ===========================
@teras_export("teras.activations.geglu")
def geglu(logits, axis: int = -1):
    """
    GeGLU is an activation function which is a variant of GLU

    Args:
        logits: `Tensor`, tensor of logits.
        axis: `int`, axis along which to apply geglu activation.
            Defaults to -1.
    """
    x, gates = ops.split(logits,
                         indices_or_sections=2,
                         axis=axis)
    return x * ops.gelu(gates)


# =========================== SPARSEMAX ===========================
@teras_export("teras.activations.sparsemax")
def sparsemax(logits, axis: int = -1):
    """
    Sparsemax activation function as proposed by T. Martins et al. in
    the paper, "From Softmax to Sparsemax: A Sparse Model of Attention
    and Multi-Label Classification"

    Reference(s):
        https://arxiv.org/abs/1602.02068

    Args:
        logits: `Tensor`, tensor of logits.
        axis: `int`, axis along which to apply the sparsemax activation.
            Defaults to -1.
    """
    K = ops.shape(logits)[-1]
    idx = ops.expand_dims(ops.arange(1, K+1, dtype=logits.dtype), 0)
    z_sorted, _ = ops.top_k(logits, k=K)
    z_cumsum = ops.cumsum(z_sorted, axis=axis, dtype=logits.dtype)
    kz = ops.sum(1 + (idx * z_sorted) > z_cumsum, axis=axis, keepdims=True)
    # subtract 1 from kz to bring indices in range [0, K),
    # instead of (0, K]
    selective_cumsum = ops.take_along_axis(z_cumsum, indices=kz - 1,
                                           axis=axis)

    threshold = (selective_cumsum - 1) / ops.cast(kz, selective_cumsum.dtype)
    logits_sub_threshold = logits - threshold
    p = ops.relu(logits_sub_threshold)
    return p


# =========================== Gumbel Softmax ===========================
@teras_export("teras.activations.gumbel_softmax")
def gumbel_softmax(logits,
                   temperature: float = 0.2,
                   hard: bool = False,
                   seed: int = None):
    """
    Implementation of the Gumbel Softmax activation function
    proposed by Eric Jang et al. in the paper
    Categorical Reparameterization with Gumbel-Softmax

    Reference(s):
        https://arxiv.org/abs/1611.01144

    Args:
        logits: `Tensor`
            Input tensor of logits.
        temperature: `float`, default 0.2,
            Controls the sharpness or smoothness of the resulting
            probability distribution. A higher temperature value leads
            to a smoother and more uniform probability distribution.
            Conversely, a lower temperature value makes the distribution
            concentrated around the category with the highest probability.
        hard: `bool`, default `False`,
            Whether to return soft probabilities or hard one hot vectors.
        seed: int, seed to use for random sampling.
    """
    u = random.uniform(ops.shape(logits),
                       minval=0,
                       maxval=1,
                       seed=seed)
    gumbels = -ops.log(-ops.log(u))
    perturbed_logits = (logits + gumbels) / temperature
    probabilities = ops.nn.softmax(perturbed_logits)
    if hard:
        one_hot_labels = ops.one_hot(ops.argmax(probabilities, axis=-1),
                                     ops.shape(logits)[-1])
        return one_hot_labels
    return probabilities



import os

import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase

from teras._src.models.backbones.tabnet.encoder import TabNetEncoderBackbone


class TabNetEncoderBackboneTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.input_batch = random.normal((self.batch_size, 5))

    # def test_valid_call(self):
    #     model = TabNetEncoderBackbone(input_dim=5,
    #                                   feature_transformer_dim=16,
    #                                   decision_step_dim=8)
    #     outputs = model(self.input_batch)
    #
    # def test_valid_output_shape(self):
    #     model = TabNetEncoderBackbone(input_dim=5,
    #                                   feature_transformer_dim=16,
    #                                   decision_step_dim=8)
    #     outputs = model(self.input_batch)
    #     self.assertEqual(ops.shape(outputs), (self.batch_size, 8))

    def test_model_save_and_load(self):
        model = TabNetEncoderBackbone(input_dim=5,
                                      feature_transformer_dim=16,
                                      decision_step_dim=8)
        outputs = model(self.input_batch)
        save_path = os.path.join(self.get_temp_dir(),
                                 "tabnet_encoder_backbone.keras"
                                 )
        model.save(save_path)
        print("Reloading model...")
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TabNetEncoderBackbone)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.input_batch)
        self.assertAllClose(
            ops.convert_to_numpy(outputs),
            ops.convert_to_numpy(reloaded_outputs)
        )




import keras

from teras._src.layers.layer_list import LayerList
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.PCGAINClassifier")
class PCGAINClassifier(keras.Model):
    """
    The auxiliary classifier for the PC-GAIN architecture proposed by Yufeng
    Wang et al. in the paper
    "PC-GAIN: Pseudo-label Conditional Generative Adversaria Imputation
    Networks for Incomplete Data"

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        num_classes: int, Number of classes to predict. It should be equal to
            the `num_clusters`, computed during the pseudo label generation.
        data_dim: int, The dimensionality of the input dataset.
            Note the dimensionality must be equal to the dimensionality of
            dataset that is passed to the fit method and not necessarily the
            dimensionality of the raw input dataset as sometimes data
            transformation alters the dimensionality of the dataset.
        hidden_dims: list, A list/tuple of units to construct hidden block of
            classifier. For each element, a new hidden layer will be added.
            By default, `hidden_dims` = [`data_dim`, `data_dim`]
        activation_hidden: Activation function to use for the hidden layers
            in the classifier. Defaults to "relu".
        activation_out: Activation function to use for the output layer of
            classifier. Defaults to "softmax".
    """
    def __init__(self,
                 num_classes: int,
                 data_dim: int,
                 hidden_dims: IntegerSequence = None,
                 activation_hidden: str = "relu",
                 activation_out: str = "softmax",
                 **kwargs):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.data_dim = data_dim
        self.hidden_dims = hidden_dims
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out

        if self.hidden_dims is None:
            self.hidden_dims = [data_dim] * 2

        self.hidden_block = []
        for i, dim in enumerate(self.hidden_dims):
            self.hidden_block.append(
                keras.layers.Dense(
                    dim,
                    activation=self.activation_hidden,
                    name=f"pcgain_classifier_hidden_layer_{i}")
            )
        self.hidden_block = LayerList(self.hidden_block,
                                      sequential=True,
                                      name="pcgain_classifier_hidden_block")
        self.output_layer = keras.layers.Dense(
            self.num_classes,
            activation=self.activation_out,
            name="pcgain_classifier_output_layer")

    def build(self, input_shape):
        self.hidden_block.build(input_shape)
        input_shape = self.hidden_block.compute_output_shape(input_shape)
        self.output_layer.build(input_shape)

    def call(self, inputs):
        x = self.hidden_block(inputs)
        return self.output_layer(x)

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.num_classes,)

    def get_config(self):
        config = super().get_config()
        config.update({
            'num_classes': self.num_classes,
            'data_dim': self.data_dim,
            'hidden_dims': self.hidden_dims,
            'activation_hidden': self.activation_hidden,
            'activation_out': self.activation_out
        })
        return config



from teras._src.layers.cls_token_extraction import CLSTokenExtraction
from keras import random, ops
from keras.src.testing.test_case import TestCase


class CLSTokenExtractionTest(TestCase):
    def setUp(self):
        self.input_shape = (8, 3, 6)
        self.input_batch = random.normal(self.input_shape)

    def test_valid_output_shape(self):
        extraction = CLSTokenExtraction()
        outputs = extraction(self.input_batch)
        self.assertEqual(ops.shape(outputs),
                         (self.input_shape[0],
                          1,
                          self.input_shape[2]))



from keras import ops
from keras import random
from sklearn.preprocessing import OneHotEncoder
from sklearn.mixture import BayesianGaussianMixture
import pandas as pd
from teras._src.preprocessing.data_transformers.data_transformer import DataTransformer
from teras._src.typing import FeaturesNamesType
from teras._src.decorators import assert_fitted
import concurrent.futures
import numpy as np
import json
import pickle


def continuous_feature_transformer(args):
    x, feature_name, weight_threshold, bgm_kwargs = args
    if isinstance(x, pd.DataFrame):
        feature = x[feature_name].values.reshape(-1, 1)
    elif isinstance(x, np.ndarray):
        feature = x[:, feature_name].reshape(-1, 1)
    else:
        raise ValueError(
            f"`x` must be either a pandas DataFrame or numpy ndarray. "
            f"{type(x)} was given.")

    feature_metadata = {}
    feature_metadata["name"] = feature_name
    bay_guass_mix = BayesianGaussianMixture(**bgm_kwargs)
    bay_guass_mix.fit(feature)
    # The authors use a weight threshold to filter out components in their
    # implementation.
    # For consistency's sake, we're going to use this idea but with slight
    # modification.
    # Reference to the official implementation:
    # https://github.com/sdv-dev/CTGAN
    valid_clusters_indicator = bay_guass_mix.weights_ > weight_threshold
    # Compute probability of coming from each cluster for each value in the
    # given (continuous) feature.
    clusters_probs = bay_guass_mix.predict_proba(feature)
    # Filter out the "invalid" clusters
    clusters_probs = clusters_probs[:, valid_clusters_indicator]
    clusters_probs[np.isnan(clusters_probs)] = 1e-10
    # Normalize probabilities to sum up to 1 for each row
    clusters_probs /= np.sum(clusters_probs, axis=1, keepdims=True)
    selected_clusters_indices = np.apply_along_axis(
        lambda cpa: np.random.choice(np.arange(len(cpa)),
                                     replace=False, p=cpa),
        axis=1,
        arr=clusters_probs)
    # To create one-hot component, we'll store the selected clusters indices
    # and the number of valid clusters
    num_valid_clusters = sum(valid_clusters_indicator)
    feature_metadata[
        'selected_clusters_indices'] = selected_clusters_indices
    feature_metadata['num_valid_clusters'] = num_valid_clusters
    # Use the selected clusters to normalize the values.
    # To normalize, we need the means and standard deviations
    # Means
    feature_metadata['clusters_means'] = bay_guass_mix.means_.squeeze()[
        valid_clusters_indicator]
    # Standard Deviations
    feature_metadata['clusters_stds'] = np.sqrt(
        bay_guass_mix.covariances_).squeeze()[valid_clusters_indicator]
    return feature_metadata


class ModeSpecificNormalization:
    """
    Mode Specific Normalization as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        continuous_features: list, List of continuous features names.
            In the case of ndarray, pass a list of continuous column indices
        max_clusters: int, Maximum clusters. Defaults to 10.
        std_multiplier: int, Multiplies the standard deviation in the
            normalization. Defaults to 4 as proposed in the paper.
        weight_threshold: float, Taken from the official implementation.
            The minimum value a component weight can take to be considered
            a valid component. `weights_` under this value will be ignored.
            Defaults to 0.005
        covariance_type: str, Parameter for the `GaussianMixtureModel`
            class of sklearn
        weight_concentration_prior_type: str, Parameter for the
            `GaussianMixtureModel` class of sklearn. Defaults to
            "dirichlet_process".
        weight_concentration_prior: float, Parameter for the
        GaussianMixtureModel class of sklearn. Defaults to 0.001.
    """
    def __init__(self,
                 continuous_features: FeaturesNamesType,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.005,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001):
        self.continuous_features = continuous_features
        self.max_clusters = max_clusters
        self.std_multiplier = std_multiplier
        self.weight_threshold = weight_threshold
        self.covariance_type = covariance_type
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior

        # Features meta-data dictionary will contain all the information
        # about a feature such as selected clusters indices, number of
        # valid clusters, clusters means & stds.
        # 1. `Clusters indices` will be used in the transform method
        #       to create the one hot vector B(i,j) where B stands for
        #       Beta, for each value c(i,j) where c(i,j) is the jth
        #       value in the ith continuous feature as proposed in the
        #       paper in the steps to apply mode specific normalization
        #       method on page 3-4.
        # 2. `Number of valid clusters` will be used in the transform
        #       method when one-hotting
        # 3. `means` and `standard deviations` will be used in transform
        #       step to normalize the value c(i,j) to create a(i,j) where
        #       a stands for alpha.
        self._metadata = {}

        # self.bay_guass_mix = BayesianGaussianMixture(
        #     n_components=self.max_clusters,
        #     covariance_type=self.covariance_type,
        #     weight_concentration_prior_type=self.weight_concentration_prior_type,
        #     weight_concentration_prior=self.weight_concentration_prior)
        self._bgm_kwargs = dict(
            n_components=self.max_clusters,
            covariance_type=self.covariance_type,
            weight_concentration_prior_type=self.weight_concentration_prior_type,
            weight_concentration_prior=self.weight_concentration_prior,
        )

        self._fitted = False

    @property
    @assert_fitted
    def metadata(self):
        return self._metadata

    def fit(self, x):
        """
        Args:
            x: A pandas DataFrame or numpy ndarray
        """
        relative_indices_all = []
        num_valid_clusters_all = []
        relative_index = 0

        feat_args = [(x, feature_name, self.weight_threshold,
                      self._bgm_kwargs)
                     for feature_name in self.continuous_features]
        with concurrent.futures.ProcessPoolExecutor() as executor:
            results = executor.map(continuous_feature_transformer,
                                   feat_args)
            for r in results:
                _name = r["name"]
                relative_indices_all.append(relative_index)
                relative_index += (1 + r["num_valid_clusters"])
                num_valid_clusters_all.append(r["num_valid_clusters"])
                del r["name"]
                self._metadata[_name] = r
        self._metadata["relative_indices_all"] = np.array(relative_indices_all)
        self._metadata["num_valid_clusters_all"] = num_valid_clusters_all
        self._fitted = True

    @assert_fitted
    def transform(self, x):
        """
        Args:
            x: A pandas DataFrame
        Returns:
            A numpy ndarray of transformed continuous data
        """
        # Contain the normalized continuous features
        x_cont_normalized = []
        for feature_name in self.continuous_features:
            selected_clusters_indices = self._metadata[feature_name][
                'selected_clusters_indices']
            num_valid_clusters = self._metadata[feature_name][
                'num_valid_clusters']
            # One hot components for all values in the feature.
            # We borrow the beta notation from the paper for clarity and
            # understanding's sake.
            betas = np.eye(num_valid_clusters)[selected_clusters_indices]

            # Normalizing
            means = self._metadata[feature_name]['clusters_means']
            stds = self._metadata[feature_name]['clusters_stds']
            if isinstance(x, pd.DataFrame):
                feature = x[feature_name].values
            elif isinstance(x, np.ndarray):
                feature = x[:, feature_name]
            else:
                raise ValueError(f"`x` must be either a pandas DataFrame "
                                 f"or numpy ndarray. {type(x)} was given.")
            means = means[selected_clusters_indices]
            stds = stds[selected_clusters_indices]
            alphas = (feature - means) / (self.std_multiplier * stds)
            alphas = ops.expand_dims(alphas, 1)

            normalized_feature = ops.concatenate([alphas, betas], axis=1)
            x_cont_normalized.append(normalized_feature)

        x_cont_normalized = ops.concatenate(x_cont_normalized, axis=1)
        return x_cont_normalized

    def fit_transform(self, x):
        """
        Fits and transforms x.
        Returns:
            A normalized copy of x.
        """
        self.fit(x)
        return self.transform(x)

    @assert_fitted
    def reverse_transform(self, x_normalized):
        """
        Args:
            x: DataFrame or n-dimensional numpy array
        Returns:
            Convert the normalized features to original values,
            effectively reversing the transformation
        """
        x = x_normalized.copy()
        for feature_name in self.continuous_features:
            means = self._metadata[feature_name]['clusters_means']
            stds = self._metadata[feature_name]['clusters_stds']
            if isinstance(x_normalized, pd.DataFrame):
                normalized_feature = x_normalized[feature_name].values
                x[feature_name] = normalized_feature * (self.std_multiplier
                                                        * stds) + means
            else:
                raise ValueError(
                    f"`x_normalized` must be either a pandas DataFrame"
                    f"Received {type(x_normalized)}")
        return x

    @classmethod
    def load_from_preset(cls, metadata, **kwargs):
        c = cls(**kwargs)
        c._metadata = metadata
        return c


class CTGANDataTransformer(DataTransformer):
    """
    Data Transformation class based on the data transformation
    in the official CTGAN paper and implementation.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        categorical_features: list, List of categorical features names in the
            dataset.
        continuous_features: list, List of continuous features names in the
            dataset.
        max_clusters: int, Maximum Number of clusters to use in
            `ModeSpecificNormalization`. Defaults to 10.
        std_multiplier: int, Multiplies the standard deviation in the
            normalization. Defaults to 4.
        weight_threshold: float, The minimum value a component weight can 
            take to be considered a valid component. `weights_` under this 
            value will be ignored. (Taken from the official implementation.)
            Defaults to 0.005.
        covariance_type: str, Parameter for the `GaussianMixtureModel` 
            class of sklearn. Defaults to "full".
        weight_concentration_prior_type: str, Parameter for the 
            `GaussianMixtureModel` class of sklearn. Defaults to 
            "dirichlet_process"
        weight_concentration_prior: float, Parameter for the 
            `GaussianMixtureModel` class of sklearn. Defaults to 0.001.
    """
    def __init__(self,
                 continuous_features: FeaturesNamesType = None,
                 categorical_features: FeaturesNamesType = None,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.005,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001
                 ):
        super().__init__()
        self.continuous_features = continuous_features if continuous_features else []
        self.categorical_features = categorical_features if categorical_features else []
        self.max_clusters = max_clusters
        self.std_multiplier = std_multiplier
        self.weight_threshold = weight_threshold
        self.covariance_type = covariance_type
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior

        self._num_categorical_features = len(categorical_features)
        self._num_continuous_features = len(continuous_features)
        self._metadata["num_categorical"] = self._num_categorical_features
        self._metadata["num_continuous"] = self._num_continuous_features

        self.mode_specific_normalizer = None
        if self._num_continuous_features > 0:
            self.mode_specific_normalizer = ModeSpecificNormalization(
                continuous_features=self.continuous_features,
                max_clusters=self.max_clusters,
                std_multiplier=self.std_multiplier,
                weight_threshold=self.weight_threshold,
                covariance_type=self.covariance_type,
                weight_concentration_prior_type=self.weight_concentration_prior_type,
                weight_concentration_prior=self.weight_concentration_prior)
        self._one_hot_enc = OneHotEncoder()
        self._fitted = False

    def fit(self, x):
        if not isinstance(x, pd.DataFrame):
            raise ValueError(
                f"`x` must be a pandas dataframe. Received {type(x)}")
        self._fit_continuous(x[self.continuous_features])
        self._fit_categorical(x[self.categorical_features])
        self._fitted = True

    def _fit_continuous(self, x_cont):
        self.mode_specific_normalizer.fit(x_cont)
        self._metadata["continuous"] = self.mode_specific_normalizer.metadata

    def _transform_continuous(self, x_cont):
        return self.mode_specific_normalizer.transform(x_cont)

    def _fit_categorical(self, x_cat):
        # To speedup computation of conditional vector down the road,
        # we assign a relative index to each feature. For instance,
        # Given three categorical columns Gender(2 categories),
        # City (4 categories) and EconomicClass (5 categories)
        # Relative indexes will be calculated as below:
        #   gender_relative_index: 0
        #   city_relative_index: gender_relative_index + num_categories_in_gender => 0 + 2 = 2
        #   economic_class_relative_index: city_relative_index + num_categories_in_city => 2 + 4 = 6

        categorical_metadata = dict()
        # NOTE: The purpose of using these lists is that, this way we'll later
        # be able to access metadata for multiple features at once using
        # their indices rather than names which would've been required in
        # case of a dict and would have been less efficient?
        relative_indices_all = []
        num_categories_all = []
        # For every feature we'll compute the probabilities over the range of
        # values based on their freqs and then append that probabilities
        # array to the following mother array
        categories_probs_all = []
        # A nested list where each element corresponds to the list of
        # categories in the feature
        categories_all = []
        feature_relative_index = 0
        for feature_name in self.categorical_features:
            num_categories = x_cat[feature_name].nunique()
            num_categories_all.append(num_categories)
            relative_indices_all.append(feature_relative_index)
            feature_relative_index += num_categories

            log_freqs = x_cat[feature_name].value_counts().apply(ops.log)
            categories_probs_dict = log_freqs.to_dict()
            # To overcome the floating point precision issue which causes
            # probabilities to not sum up to 1 and resultantly causes error
            # in ops.random.choice method, we round the probabilities to 7
            # decimal points
            probs = ops.round(ops.array(list(categories_probs_dict.values())),
                              decimals=7)
            # Normalizing so all probs sum up to 1
            probs = probs / ops.sum(probs)
            categories, categories_probs = list(categories_probs_dict.keys()), probs
            categories_probs_all.append(categories_probs)
            categories_all.append(categories)
        categorical_metadata["total_num_categories"] = sum(num_categories_all)
        categorical_metadata["num_categories_all"] = num_categories_all
        categorical_metadata["relative_indices_all"] = ops.array(
            relative_indices_all)
        categorical_metadata["categories_probs_all"] = categories_probs_all
        categorical_metadata["categories_all"] = categories_all

        self._metadata["categorical"] = categorical_metadata
        self._one_hot_enc.fit(x_cat)

    def _transform_categorical(self, x_cat):
        return self._one_hot_enc.transform(x_cat)

    @assert_fitted
    def transform(self, x):
        total_transformed_features = 0
        x_continuous, x_categorical = None, None
        if self._num_continuous_features > 0:
            x_continuous = self._transform_continuous(
                x[self.continuous_features])
            total_transformed_features += (
                    self._metadata["continuous"]["relative_indices_all"][-1] +
                    self._metadata["continuous"]["num_valid_clusters_all"][-1])
        if self._num_categorical_features > 0:
            x_categorical = self._transform_categorical(
                x[self.categorical_features])
            total_transformed_features += (
                    self._metadata["categorical"]["relative_indices_all"][-1] +
                    self._metadata["categorical"]["num_categories_all"][-1] + 1)

        # since we concatenate the categorical features AFTER the continuous
        # alphas and betas so we'll create an overall relative indices array
        # where we offset the relative indices of the categorical features by
        # the total number of continuous features components
        relative_indices_all = []
        offset = 0
        if x_continuous is not None:
            cont_relative_indices = self._metadata["continuous"]["relative_indices_all"]
            relative_indices_all.extend(cont_relative_indices)
            offset = cont_relative_indices[-1] + self._metadata["continuous"]["num_valid_clusters_all"][-1]
        if x_categorical is not None:
            # +1 since the categorical relative indices start at 0
            relative_indices_all.extend(self._metadata["categorical"]["relative_indices_all"] + 1 + offset)
        self._metadata["relative_indices_all"] = relative_indices_all
        self._metadata["total_transformed_features"] = total_transformed_features
        x_transformed = ops.concatenate([x_continuous, x_categorical.toarray()], axis=1)
        return x_transformed

    def reverse_transform(self, x_generated):
        """
        Reverses transforms the generated data to the original data format.

        Args:
            x_generated: Generated dataset.

        Returns:
            Generated data in the original data format.
        """
        all_features = self.continuous_features + self.categorical_features
        if self._num_continuous_features > 0:
            num_valid_clusters_all = self._metadata["continuous"]["num_valid_clusters_all"]
        if self._num_categorical_features > 0:
            num_categories_all = self._metadata["categorical"]["num_categories_all"]
        data = {}
        cat_index = 0       # categorical index
        cont_index = 0      # continuous index
        for index, feature_name in enumerate(all_features):
            # the first n features are continuous
            if index < len(self.continuous_features):
                alphas = x_generated[:, index]
                betas = x_generated[:, index + 1 : index + 1 + num_valid_clusters_all[cont_index]]
                # Recall that betas represent the one hot encoded form of the
                # cluster number
                cluster_indices = ops.argmax(betas, axis=1)
                # List of cluster means for a feature. contains one value per
                # cluster
                means = self._metadata["continuous"][feature_name]["clusters_means"]

                # Since each individual element within the cluster is associated
                # with one of the cluster's mean. We use the
                # `cluster_indices` to get a list of size len(x) where each
                # element is a mean for the corresponding element in the feature
                means = means[cluster_indices]
                # Do the same for stds
                stds = self._metadata["continuous"][feature_name]["clusters_stds"]
                stds = stds[cluster_indices]
                feature = alphas * (self.std_multiplier * stds) + means
                data[feature_name] = feature
                cont_index += 1
            else:
                # if the index is greater than or equal to
                # len(continuous_features), then the column at hand is
                # categorical
                raw_feature_parts = x_generated[:, index: index + num_categories_all[cat_index]]
                categories = self._one_hot_enc.categories_[cat_index]
                categories_indices = ops.argmax(raw_feature_parts, axis=1)
                feature = categories[categories_indices]
                data[feature_name] = feature
                cat_index += 1
        return pd.DataFrame(data)

    def save(self, filename):
        """
        Saves the fitted state of `CTGANDataTransformer` instance for
        portability, in the `json` format.
        It also saves a binary file with same name (excluding the file
        extension ofcourse.)

        Args:
            filename: Filename or file path ending in `.json` extension.
        """
        args = {
            "categorical_features": self.categorical_features,
            "continuous_features": self.continuous_features,
            "max_clusters": self.max_clusters,
            "std_multiplier": self.std_multiplier,
            "weight_threshold": self.weight_threshold,
            "covariance_type": self.covariance_type,
            "weight_concentration_prior_type":
                self.weight_concentration_prior_type,
            "weight_concentration_prior": self.weight_concentration_prior,
        }
        attrs = {
            "_metadata": self._metadata,
            "_fitted": self._fitted
        }
        state = {
            "args": args,
            "attrs": attrs,
        }
        with open(filename, "w") as f:
            json.dump(state, f)
        encoder_filename = str(filename).replace(".json", "_encoder.pkl")
        with open(encoder_filename, "wb") as f:
            pickle.dump(self._one_hot_enc)

    @classmethod
    def load(cls, filename):
        """
        Loads the saved state of `CTGANDataTransformer` from the `json` file.
        It also loads the pickled `OneHotEncoder` instance.

        Args:
            filename: Filename or file path ending in `.json` extension.

        Returns:
            An instance of `CTGANDataTransformer` with state stored in the
            `filename` json file.
        """
        with open(filename, "r") as f:
            state = json.load(f)
        c = cls(**state.pop("params"))
        for name, value in state.pop("attrs"):
            c.__setattr__(name, value)
        encoder_filename = str(filename).replace(".json", "_encoder.pkl")
        with open(encoder_filename, "rb") as f:
            c._one_hot_enc = pickle.load(f)
        return c



import keras
from keras import ops

from teras._src.layers.layer_list import LayerList
from teras._src.layers.tabnet.feature_transformer_layer import \
    TabNetFeatureTransformerLayer
from teras._src.api_export import teras_export


@teras_export("teras.models.TabNetDecoder")
class TabNetDecoder(keras.Model):
    """
    TabNetDecoder model for self-supervised learning,
    proposed by Arik et al. in the
    "TabNet: Attentive Interpretable Tabular Learning" paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        data_dim: int, The dimensionality of the input dataset.
        feature_transformer_dim: int, the dimensionality of the hidden
            representation in feature transformation block.
            The dense layer inside the `TabNetFeatureTransformer` first
            maps  the inputs to `feature_transformer_dim` * 2 dimension
            hidden representations and later the glu activation
            maps the hidden representations to
            `feature_transformer_dim` dimensions.
        decision_step_dim: int, the dimensionality of output at each
            decision step, which is later mapped to the final
            classification or regression output.
            It is recommended to keep `decision_step_dim` and
            `feature_transformer_dim` equal to each other.
            Adjusting these two parameters values is a good way of
            obtaining a tradeoff between performance and complexity.
        num_decision_steps: int, the number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset,
            the value tends to be higher. That said, a very high value
            of `num_decision_steps` may suffer from overfitting.
            Defaults to 5.
        num_shared_layers: int, Number of shared layers to use in the
            ``TabNetFeatureTransformer``.
            These shared layers are shared across decision steps.
            Defaults to 2.
        num_decision_dependent_layers: int, number of decision dependent
            layers to use in the `TabNetFeatureTransformer`.
            In simple words, `num_decision_dependent_layers` are created
            for each decision step in the `num_decision_steps`.
            For instance, if `num_decision_steps = 5` and
            `num_decision_dependent_layers = 2`
            then 10 layers will be created, 2 for each decision step.
        relaxation_factor: float, relaxation factor that promotes the
            reuse of each feature at different decision steps. When it
            is 1, a feature is enforced to be used only at one decision
            step and as it increases, more flexibility is provided to
            use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on
            the performance. Typically, a larger value for
            `num_decision_steps` favors for a larger `relaxation_factor`.
        batch_momentum: float, momentum value to use for
            `BatchNormalization` layer.
            Defaults to 0.9
        epsilon: float, epsilon is a small number for numerical stability
            during the computation of entropy loss.
            Defaults to 0.00001
    """
    def __init__(self,
                 data_dim: int,
                 feature_transformer_dim: int,
                 decision_step_dim: int,
                 num_decision_steps: int = 5,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 5,
                 relaxation_factor: float = 1.5,
                 batch_momentum: float = 0.9,
                 epsilon: float = 1e-5,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.feature_transformer_dim = feature_transformer_dim
        self.decision_step_dim = decision_step_dim
        self.num_decision_steps = num_decision_steps
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.relaxation_factor = relaxation_factor
        self.batch_momentum = batch_momentum
        self.epsilon = epsilon

        # self.feature_transformers = [
        #     TabNetFeatureTransformer(
        #         hidden_dim=self.feature_transformer_dim,
        #         num_shared_layers=self.num_decision_steps,
        #         num_decision_dependent_layers=self.num_decision_dependent_layers,
        #         batch_momentum=self.batch_momentum,
        #         name=f"decoder_feature_transformer_{i}"
        #     )
        #     for i in range(self.num_decision_steps)
        # ]
        self.shared_layers = LayerList([
            TabNetFeatureTransformerLayer(
                dim=feature_transformer_dim,
                batch_momentum=batch_momentum,
                name=f"decoder_shared_layer_{i}")
            for i in range(self.num_shared_layers)
        ],
            name="shared_layers")
        self.decision_dependent_layers = LayerList([
            TabNetFeatureTransformerLayer(
                dim=feature_transformer_dim,
                batch_momentum=batch_momentum,
                name=f"decoder_decision_dependent_layer_{i}")
            for i in range(self.num_decision_steps * self.num_decision_dependent_layers)
        ],
            name="decision_dependent_layers"
        )
        self.projection_layers = LayerList([
            keras.layers.Dense(units=self.data_dim,
                               name=f"projection_layer_{i}")
            for i in range(self.num_decision_steps)
        ],
            sequential=False,
            name="projection_layers"
        )

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self._input_shape = input_shape
        if not self.shared_layers.built:
            self.shared_layers.build(input_shape)
        input_shape = self.shared_layers.compute_output_shape(input_shape)
        self.decision_dependent_layers.build(input_shape)
        input_shape = self.decision_dependent_layers.compute_output_shape(
            input_shape
        )
        self.projection_layers.build(input_shape)

    def call(self, inputs, mask=None):
        batch_size = ops.shape(inputs)[0]
        reconstructed_features = ops.zeros(
            shape=(batch_size, self.data_dim)
        )

        for i in range(self.num_decision_steps):
            # feature_out = self.feature_transformers[i](inputs)
            # reconstructed_features += self.projection_layers[i](feature_out)
            x = inputs
            residue = None
            for layer in self.shared_layers:
                x = layer(x)
                if residue is not None:
                    x += ops.sqrt(0.5) * residue
                residue = x

            start_idx = i * self.num_decision_steps
            end_idx = start_idx + self.num_decision_steps
            for layer in self.decision_dependent_layers[start_idx: end_idx]:
                x = layer(x)
                if residue is not None:
                    x += ops.sqrt(0.5) * residue
                residue = x
            reconstructed_features += self.projection_layers[i](x)

        # According to the paper, the decoder’s last FC (dense) layer is
        # multiplied with S (binary mask indicating which features are
        # missing) to output the unknown features.
        if mask is not None:
            reconstructed_features *= mask

        return reconstructed_features

    def get_config(self):
        config = super().get_config()
        config.update({
            'data_dim': self.data_dim,
            'feature_transformer_dim': self.feature_transformer_dim,
            'decision_step_dim': self.decision_step_dim,
            'num_decision_steps': self.num_decision_steps,
            'num_shared_layers': self.num_shared_layers,
            'num_decision_dependent_layers': self.num_decision_dependent_layers,
            'relaxation_factor': self.relaxation_factor,
            'batch_momentum': self.batch_momentum,
            'epsilon': self.epsilon,
        })
        return config

    def get_build_config(self):
        build_config = dict(input_shape=self._input_shape)
        return build_config

    def build_from_config(self, build_config):
        self.build(build_config["input_shape"])



import keras
from teras._src.activations import gumbel_softmax
from teras._src.api_export import teras_export


@teras_export("teras.layers.GumbelSoftmax")
class GumbelSoftmax(keras.layers.Layer):
    """
    Implementation of the Gumbel Softmax activation
    proposed by Eric Jang et al. in the paper,
    "Categorical Reparameterization with Gumbel-Softmax"

    Reference(s):
        https://arxiv.org/abs/1611.01144

    Args:
        temperature: float, Controls the sharpness or smoothness of the
            resulting probability distribution. A higher temperature value
            leads to a smoother and more uniform probability distribution.
            Conversely, a lower temperature value makes the distribution
            concentrated around the category with the highest probability.
        hard: bool, Whether to return soft probabilities or hard one hot
        vectors. Defaults to False.
        seed: int, Seed for random sampling.
    """
    def __init__(self,
                 temperature: float = 0.2,
                 hard: bool = False,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature
        self.hard = hard
        self.seed = seed

    def build(self, input_shape):
        # nothing to build
        self.built = True

    def call(self, logits):
        return gumbel_softmax(logits,
                              temperature=self.temperature,
                              hard=self.hard,
                              seed=self.seed)

    def get_config(self):
        config = super().get_config()
        new_config = {'temperature': self.temperature,
                      'hard': self.hard}
        config.update(new_config)
        return config



import keras
from keras import ops
from teras._src.layers.layer_list import LayerList
from teras._src.api_export import teras_export


@teras_export("teras.layers.SAINTReconstructionHead")
class SAINTReconstructionHead(keras.layers.Layer):
    """
    SAINT Reconstruction Head layer for `SAINTPretrainer`.
    For each feature in the dataset, it creates an MLP with a hidden
    layer and an output layer with dimensions equal to the cardinality of
    the feature in the case of a categorical features and 1 in the case
    of a continuous feature.

    Args:
        cardinalities: list or ndarray, a list or 1d-array of
            cardinalities of all the features in the dataset in the
            same order as the features' occurrence.
            For numerical features, use 0 as indicator at
            the corresponding index of the array.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, Dimensionality of embeddings being used in
            the model,
    """
    def __init__(self,
                 cardinalities: list,
                 embedding_dim: int,
                 **kwargs):
        super().__init__(**kwargs)
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim

        self.reconstruction_blocks = LayerList([
            LayerList([
                keras.layers.Dense(embedding_dim * 5,
                                   activation="relu",
                                   name=f"hidden_rb_{i}"),
                keras.layers.Dense((card + 1) if card == 0 else card)],
                sequential=True,
                name=f"reconstruction_block_{i}"
            )
            for i, card in enumerate(self.cardinalities)
        ],
            sequential=False,
            name="reconstruction_blocks"
        )

    def build(self, input_shape):
        self.reconstruction_blocks.build(input_shape)

    def call(self, inputs):
        reconstructed_features = self.reconstruction_blocks[0](
            inputs[:, 0, :]
        )
        for idx, layer in enumerate(self.reconstruction_blocks[1:],
                                    start=1):
            r_f = layer(inputs[:, idx, :])
            reconstructed_features = ops.concatenate(
                [reconstructed_features, r_f],
                axis=1
            )
        return reconstructed_features

    def compute_output_shape(self, input_shape):
        return input_shape[:1] + (len(self.cardinalities),)

    def get_config(self):
        config = super().get_config()
        config.update({
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim,
        })
        return config



import keras
from keras import random, ops
from keras.backend import floatx
from keras.backend import backend


class BaseGAIN(keras.Model):
    """
    Base class for GAIN.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 100.,
                 seed: int = 1337,
                 **kwargs):
        if not backend() == "jax":
             # Don't call super() with JAX backend as `JAXGAN` does that
             # already!
             super().__init__(**kwargs)
        self.generator = generator
        self.discriminator = discriminator
        self.hint_rate = hint_rate
        self.alpha = alpha
        self.seed = seed

        # Loss trackers
        self.generator_loss_tracker = keras.metrics.Mean(
            name="generator_loss")
        self.discriminator_loss_tracker = keras.metrics.Mean(
            name="discriminator_loss")

        self._seed_gen = random.SeedGenerator(self.seed)

    def build(self, input_shape):
        # Inputs received by each generator and discriminator have twice the
        # dimensions of original inputs
        input_shape = (input_shape[:-1], input_shape[-1] * 2)
        self.generator.build(input_shape)
        self.discriminator.build(input_shape)

    def compile(self,
                generator_optimizer=keras.optimizers.Adam(),
                discriminator_optimizer=keras.optimizers.Adam(),
                **kwargs
                ):
        super().compile(**kwargs)
        self.generator_optimizer = generator_optimizer
        self.discriminator_optimizer = discriminator_optimizer

    @property
    def metrics(self):
        metrics = [
                   self.generator_loss_tracker,
                   self.discriminator_loss_tracker,
                   ]
        return metrics

    def compute_loss(self, **kwargs):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` doesn't provide an implementation for"
            f" the `compute_loss` method. Please use "
            f"`compute_discriminator_loss` or `compute_generator_loss` for "
            f"relevant purpose."
        )

    def compute_discriminator_loss(self, mask, mask_pred):
        return keras.losses.BinaryCrossentropy()(mask, mask_pred)

    def compute_generator_loss(self, x, x_generated, mask, mask_pred, alpha):
        cross_entropy_loss = keras.losses.CategoricalCrossentropy()(
            mask, mask_pred
        )
        mse_loss = keras.losses.MeanSquaredError()(
            y_true=(mask * x),
            y_pred=(mask * x_generated))
        loss = cross_entropy_loss + alpha * mse_loss
        return loss

    def call(self, **kwargs):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` doesn't provide an implementation "
            f"for the `call` method. Please use the call method of "
            f"`GAIN().generator` or `GAIN().discriminator`."
        )

    def test_step(self, data):
        x_gen, x_disc = data

        # =========================
        # Test the discriminator
        # =========================
        # Create mask
        mask = 1. - ops.cast(ops.isnan(x_disc), dtype=floatx())
        # replace nans with 0.
        x_disc = ops.where(ops.isnan(x_disc), x1=0., x2=x_disc)
        # Sample noise
        z = random.uniform(shape=ops.shape(x_disc), minval=0.,
                           maxval=0.01, seed=self._seed_gen)
        # Sample hint vectors
        hint_vectors = random.binomial(shape=ops.shape(x_disc),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        # Combine random vectors with original data
        x_disc = x_disc * mask + (1 - mask) * z
        x_generated = self.generator(ops.concatenate([x_disc, mask], axis=1))
        # Combine generated samples with original data
        x_hat_disc = (x_generated * (1 - mask)) + (x_disc * mask)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat_disc, hint_vectors], axis=1))
        d_loss = self.compute_discriminator_loss(mask, mask_pred)

        # =====================
        # Test the generator
        # =====================
        mask = 1. - ops.cast(ops.isnan(x_gen), dtype=floatx())
        x_gen = ops.where(ops.isnan(x_gen), x1=0., x2=x_gen)
        z = random.uniform(shape=ops.shape(x_gen), minval=0.,
                           maxval=0.01, seed=self._seed_gen)
        hint_vectors = random.binomial(shape=ops.shape(x_gen),
                                       counts=1,
                                       probabilities=self.hint_rate,
                                       seed=self._seed_gen)
        hint_vectors *= mask
        x_gen = x_gen * mask + (1 - mask) * z

        x_generated = self.generator(
            ops.concatenate([x_gen, mask], axis=1))
        # Combine generated samples with original/observed data
        x_hat = (x_generated * (1 - mask)) + (x_gen * mask)
        mask_pred = self.discriminator(
            ops.concatenate([x_hat, hint_vectors], axis=1))
        g_loss = self.compute_generator_loss(
            x=x_gen,
            x_generated=x_generated,
            mask=mask,
            mask_pred=mask_pred,
            alpha=self.alpha
        )

        # Update custom tracking metrics
        self.generator_loss_tracker.update_state(g_loss)
        self.discriminator_loss_tracker.update_state(d_loss)

        logs = {m.name: m.result() for m in self.metrics}
        return logs

    def get_generator(self):
        return self.generator

    def get_discriminator(self):
        return self.discriminator

    def predict_step(self, data):
        """
        Args:
            Transformed data.
        Returns:
            Imputed data that should be reverse transformed
            to its original form.
        """
        if isinstance(data, tuple):
            data = data[0]
        data = ops.cast(data, floatx())
        # Create mask
        mask = 1. - ops.cast(ops.isnan(data), dtype=floatx())
        data = ops.where(ops.isnan(data), x1=0., x2=data)
        # Sample noise
        z = random.uniform(ops.shape(data), minval=0.,
                           maxval=0.01, seed=self._seed_gen)
        x = mask * data + (1 - mask) * z
        imputed_data = self.generator(ops.concatenate([x, mask], axis=1))
        imputed_data = mask * data + (1 - mask) * imputed_data
        return imputed_data

    def get_config(self):
        config = super().get_config()
        config.update({
            'generator': keras.layers.serialize(self.generator),
            'discriminator': keras.layers.serialize(self.discriminator),
            'hint_rate': self.hint_rate,
            'alpha': self.alpha,
            'seed': self.seed,
        })
        return config

    @classmethod
    def from_config(cls, config):
        generator = keras.layers.deserialize(config.pop("generator"))
        discriminator = keras.layers.deserialize(config.pop("discriminator"))
        return cls(generator=generator, discriminator=discriminator,
                   **config)
